import logging
from typing import Any, Dict, Generic, List, Optional, Set, TypeVar, Union

from pie_core.utils.dictionary import flatten_dict_s
from torch import Tensor
from torchmetrics import Metric, MetricCollection

from pytorch_ie import PyTorchIEModel

from .has_taskmodule import HasTaskmodule
from .stages import TESTING, TRAINING, VALIDATION

InputType = TypeVar("InputType")
TargetType = TypeVar("TargetType")
OutputType = TypeVar("OutputType")

# from torchmetrics
_METRIC = (Metric, Tensor, int, float)

logger = logging.getLogger(__name__)


class ModelWithMetricsFromTaskModule(
    HasTaskmodule, PyTorchIEModel, Generic[InputType, TargetType, OutputType]
):
    """A PyTorchIEModel that adds metrics from a taskmodule.

    The metrics are added to the model as attributes with the names metric_{stage} via
    setup_metrics method, where stage is one of "train", "val", or "test". The metrics are updated
    with the update_metric method and logged with the log_metric method.

    Args:
        metric_stages: The stages for which to set up metrics. Must be one of "train", "val", or
            "test".
        metric_intervals: A dict mapping metric stages to the number of steps between metric
            calculation. If not provided, the metrics are calculated at the end of each epoch.
        metric_call_predict: Whether to call predict() and use its result for metric calculation
            instead of the (decoded) model output. This is useful, for instance, for generative models
            that define special logic to produce predictions, e.g. beam search, which requires multiple
            passes through the model. If True, predict() is called for all metric stages. If False (default),
            the model outputs are passed to decode() and that is used for all metric stages. If a list of
            metric stages is provided, predict() is called for these stages and the (decoded) model
            outputs for the remaining stages.
    """

    def __init__(
        self,
        metric_stages: List[str] = [TRAINING, VALIDATION, TESTING],
        metric_intervals: Optional[Dict[str, int]] = None,
        metric_call_predict: Union[bool, List[str]] = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.setup_metrics(metric_stages=metric_stages)

        self.metric_intervals = metric_intervals or {}
        missed_stages = set(self.metric_intervals) - set(metric_stages)
        if len(missed_stages) > 0:
            logger.warning(
                f"There are stages in metric_intervals that are not in metric_stages: "
                f"{missed_stages}. Available metric stages: {metric_stages}."
            )

        self.use_prediction_for_metrics: Set[str]
        if isinstance(metric_call_predict, bool):
            self.metric_call_predict = set(metric_stages) if metric_call_predict else set()
        else:
            self.metric_call_predict = set(metric_call_predict)
        missed_stages = self.metric_call_predict - set(metric_stages)
        if len(missed_stages) > 0:
            logger.warning(
                f"There are stages in metric_call_predict that are not in metric_stages: "
                f"{missed_stages}. Available metric stages: {metric_stages}."
            )

    def setup_metrics(self, metric_stages: List[str]) -> None:
        """Set up metrics for the given stages if a taskmodule is available.

        Args:
            metric_stages: The stages for which to set up metrics. Must be one of "train", "val", or
                "test".
        """
        if self.taskmodule is not None:
            for stage in metric_stages:
                metric = self.taskmodule.configure_model_metric(stage=stage)
                if metric is not None:
                    self._set_metric(stage=stage, metric=metric)
                else:
                    logger.warning(
                        f"The taskmodule {self.taskmodule.__class__.__name__} does not define a metric for stage "
                        f"'{stage}'."
                    )
        elif len(metric_stages) > 0:
            logger.warning(
                "No taskmodule is available, so no metrics are set up. "
                "Please provide a taskmodule_config to enable metrics for stages "
                f"{metric_stages}."
            )

    def _get_metric(
        self, stage: str, batch_idx: int = 0
    ) -> Optional[Union[Metric, MetricCollection]]:
        metric_interval = self.metric_intervals.get(stage, 1)
        if (batch_idx + 1) % metric_interval == 0:
            return getattr(self, f"metric_{stage}", None)
        else:
            return None

    def _set_metric(self, stage: str, metric: Optional[Union[Metric, MetricCollection]]) -> None:
        setattr(self, f"metric_{stage}", metric)

    def update_metric(
        self,
        stage: str,
        inputs: InputType,
        targets: TargetType,
        outputs: OutputType,
    ) -> None:
        """Update the metric for the given stage. If outputs is provided, the predictions are
        decoded from the outputs. Otherwise, the predictions are obtained by directly calling the
        predict method with the inputs (note that this causes the model to be called a second
        time). Finally, the metric is updated with the predictions and targets.

        Args:
            stage: The stage for which to update the metric. Must be one of "train", "val", or "test".
            inputs: The inputs to the model.
            targets: The targets for the inputs.
            outputs: The outputs of the model. They are decoded into predictions if provided. If
                outputs is None, the predictions are obtained by directly calling the predict method
                on the inputs.
        """

        metric = self._get_metric(stage=stage)
        if metric is not None:
            if stage in self.metric_call_predict:
                predictions = self.predict(inputs=inputs)
            else:
                predictions = self.decode(inputs=inputs, outputs=outputs)
            metric.update(predictions, targets)

    def log_metric(self, stage: str, reset: bool = True) -> None:
        """Log the metric for the given stage and reset it."""

        metric = self._get_metric(stage=stage)
        if metric is not None:
            values = metric.compute()
            log_kwargs: Dict[str, Any] = {"on_step": False, "on_epoch": True, "sync_dist": True}
            if isinstance(values, dict):
                values_flat = flatten_dict_s(values, sep="/")
                for key, value in values_flat.items():
                    if isinstance(value, _METRIC):
                        self.log(f"metric/{key}/{stage}", value, **log_kwargs)
                    else:
                        raise ValueError(
                            f"Metric compute() returned an unsupported type for key '{key}': {type(value)}. "
                            "Expected a single value (int, float, or Tensor) or a torchmetrics.Metric."
                        )
            elif isinstance(values, _METRIC):
                metric_name = getattr(metric, "name", None) or type(metric).__name__
                self.log(f"metric/{metric_name}/{stage}", values, **log_kwargs)
            else:
                raise ValueError(
                    f"Metric compute() returned an unsupported type: {type(values)}. "
                    "Expected a dict or a single value (int, float, or Tensor) or a torchmetrics.Metric."
                )
            if reset:
                metric.reset()



"""
workflow:
    Document
        -> (InputEncoding, TargetEncoding) -> TaskEncoding -> TaskBatchEncoding
            -> ModelBatchEncoding -> ModelBatchOutput
        -> TaskOutput
    -> Document
"""

import logging
from collections import defaultdict
from functools import partial
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    Union,
)

import numpy as np
import torch
from pie_core import Annotation, AnnotationLayer, Document, TaskEncoding, TaskModule
from pie_documents.annotations import (
    BinaryRelation,
    LabeledSpan,
    MultiLabeledBinaryRelation,
    NaryRelation,
    Span,
)
from pie_documents.documents import (
    TextBasedDocument,
    TextDocumentWithLabeledSpansAndBinaryRelations,
    TextDocumentWithLabeledSpansBinaryRelationsAndLabeledPartitions,
)
from pie_documents.utils.span import distance as span_distance
from pie_documents.utils.span import is_contained_in
from torch import LongTensor
from torchmetrics import ClasswiseWrapper, F1Score, MetricCollection
from transformers import AutoTokenizer
from transformers.file_utils import PaddingStrategy
from transformers.tokenization_utils_base import TruncationStrategy
from typing_extensions import TypeAlias, TypeVar

from pytorch_ie.models.simple_sequence_classification import InputType as ModelInputType
from pytorch_ie.models.simple_sequence_classification import TargetType as ModelTargetType
from pytorch_ie.taskmodules.interface import ChangesTokenizerVocabSize
from pytorch_ie.utils.window import get_window_around_slice

from ..utils.document import SpanNotAlignedWithTokenException, get_aligned_token_span
from .common.mixins import RelationStatisticsMixin
from .metrics import WrappedMetricWithPrepareFunction

# from pytorch_ie.utils.tokenization import (
#    SpanNotAlignedWithTokenException,
#    get_aligned_token_span,
# )

InputEncodingType: TypeAlias = Dict[str, Any]
TargetEncodingType: TypeAlias = Sequence[int]
DocumentType: TypeAlias = TextBasedDocument

TaskEncodingType: TypeAlias = TaskEncoding[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
]


class TaskOutputType(TypedDict, total=False):
    labels: Sequence[str]
    probabilities: Sequence[float]


TaskModuleType: TypeAlias = TaskModule[
    # _InputEncoding, _TargetEncoding, _TaskBatchEncoding, _ModelBatchOutput, _TaskOutput
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
    Tuple[ModelInputType, Optional[ModelTargetType]],
    ModelTargetType,
    TaskOutputType,
]


HEAD = "head"
TAIL = "tail"
START = "start"
END = "end"


logger = logging.getLogger(__name__)


def _get_labels(model_output: ModelTargetType) -> LongTensor:
    result = model_output["labels"]
    assert isinstance(result, LongTensor)
    return result


def _get_labels_together_remove_none_label(
    predictions: ModelTargetType, targets: ModelTargetType, none_idx: int
) -> Tuple[LongTensor, LongTensor]:
    mask_not_both_none = (predictions["labels"] != none_idx) | (targets["labels"] != none_idx)
    predictions_not_none = predictions["labels"][mask_not_both_none]
    targets_not_none = targets["labels"][mask_not_both_none]
    assert isinstance(predictions_not_none, LongTensor)
    assert isinstance(targets_not_none, LongTensor)
    return predictions_not_none, targets_not_none


def find_sublist(sub: List, bigger: List) -> int:
    if not bigger:
        return -1
    if not sub:
        return 0
    first, rest = sub[0], sub[1:]
    pos = 0
    try:
        while True:
            pos = bigger.index(first, pos) + 1
            if not rest or bigger[pos : pos + len(rest)] == rest:
                return pos - 1
    except ValueError:
        return -1


class MarkerFactory:
    def __init__(self, role_to_marker: Dict[str, str]):
        self.role_to_marker = role_to_marker

    def _get_role_marker(self, role: str) -> str:
        return self.role_to_marker[role]

    def _get_marker(self, role: str, is_start: bool, label: Optional[str] = None) -> str:
        result = "["
        if not is_start:
            result += "/"
        result += self._get_role_marker(role)
        if label is not None:
            result += f":{label}"
        result += "]"
        return result

    def get_start_marker(self, role: str, label: Optional[str] = None) -> str:
        return self._get_marker(role=role, is_start=True, label=label)

    def get_end_marker(self, role: str, label: Optional[str] = None) -> str:
        return self._get_marker(role=role, is_start=False, label=label)

    def get_append_marker(self, role: str, label: Optional[str] = None) -> str:
        role_marker = self._get_role_marker(role)
        if label is None:
            return f"[{role_marker}]"
        else:
            return f"[{role_marker}={label}]"

    @property
    def all_roles(self) -> Set[str]:
        return set(self.role_to_marker)

    def get_all_markers(
        self,
        entity_labels: List[str],
        append_markers: bool = False,
        add_type_to_marker: bool = False,
    ) -> List[str]:
        result: Set[str] = set()
        if add_type_to_marker:
            none_and_labels = [None] + entity_labels
        else:
            none_and_labels = [None]
        for role in self.all_roles:
            # create start and end markers without label and for all labels, if add_type_to_marker
            for maybe_label in none_and_labels:
                result.add(self.get_start_marker(role=role, label=maybe_label))
                result.add(self.get_end_marker(role=role, label=maybe_label))
            # create append markers for all labels
            if append_markers:
                for entity_label in entity_labels:
                    result.add(self.get_append_marker(role=role, label=entity_label))

        # sort and convert to list
        return sorted(result)


class RelationArgument:
    def __init__(
        self,
        entity: LabeledSpan,
        role: str,
        token_span: Span,
        add_type_to_marker: bool,
        marker_factory: MarkerFactory,
    ) -> None:
        self.marker_factory = marker_factory
        if role not in self.marker_factory.all_roles:
            raise ValueError(
                f"role='{role}' not in known roles={sorted(self.marker_factory.all_roles)} (did you "
                f"initialise the taskmodule with the correct argument_role_to_marker dictionary?)"
            )

        self.entity = entity

        self.role = role
        self.token_span = token_span
        self.add_type_to_marker = add_type_to_marker

    @property
    def maybe_label(self) -> Optional[str]:
        return self.entity.label if self.add_type_to_marker else None

    @property
    def as_start_marker(self) -> str:
        return self.marker_factory.get_start_marker(role=self.role, label=self.maybe_label)

    @property
    def as_end_marker(self) -> str:
        return self.marker_factory.get_end_marker(role=self.role, label=self.maybe_label)

    @property
    def as_append_marker(self) -> str:
        # Note: we add the label in either case (we use self.entity.label instead of self.label)
        return self.marker_factory.get_append_marker(role=self.role, label=self.entity.label)

    def shift_token_span(self, value: int):
        self.token_span = Span(
            start=self.token_span.start + value, end=self.token_span.end + value
        )


def get_relation_argument_spans_and_roles(
    relation: Annotation,
) -> Tuple[Tuple[str, Annotation], ...]:
    if isinstance(relation, BinaryRelation):
        return (HEAD, relation.head), (TAIL, relation.tail)
    elif isinstance(relation, NaryRelation):
        # create unique order by sorting the arguments by their start and end positions and role
        sorted_args = sorted(
            zip(relation.roles, relation.arguments),
            key=lambda role_and_span: (
                role_and_span[1].start,
                role_and_span[1].end,
                role_and_span[0],
            ),
        )
        return tuple(sorted_args)
    else:
        raise NotImplementedError(
            f"the taskmodule does not yet support getting relation arguments for type: {type(relation)}"
        )


def construct_mask(input_ids: torch.LongTensor, positive_ids: List[Any]) -> torch.LongTensor:
    """Construct a mask for the input_ids where all entries in mask_ids are 1."""
    masks = [torch.nonzero(input_ids == marker_token_id) for marker_token_id in positive_ids]
    global_mask = torch.cat(masks)
    value = torch.ones(global_mask.shape[0], dtype=torch.long)
    mask = torch.zeros(input_ids.shape, dtype=torch.long)
    mask.index_put_(tuple(global_mask.t()), value)
    assert isinstance(mask, torch.LongTensor)
    return mask


S = TypeVar("S", bound=Span)


def shift_span(span: S, offset: int) -> S:
    return span.copy(start=span.start + offset, end=span.end + offset)


def bio_encode_spans(
    spans: List[Tuple[int, int, str]], total_length: int, label2idx: Dict[str, int]
) -> List[int]:
    # result = ["O"] * total_length
    result = [0] * total_length
    for start, end, label in spans:
        # result[start] = f"B-{label}"
        result[start] = label2idx[label] * 2 + 1
        for i in range(start + 1, end):
            # result[i] = f"I-{label}"
            result[i] = label2idx[label] * 2 + 2
    return result


@TaskModule.register()
class RETextClassificationWithIndicesTaskModule(
    RelationStatisticsMixin,
    TaskModuleType,
    ChangesTokenizerVocabSize,
):
    """Marker based relation extraction. This taskmodule prepares the input token ids in such a way
    that before and after the candidate head and tail entities special marker tokens are inserted.
    Then, the modified token ids can be simply passed into a transformer based text classifier
    model.

    parameters:

        partition_annotation: str, optional. If specified, LabeledSpan annotations with this name are
            expected to define partitions of the document that will be processed individually, e.g. sentences
            or sections of the document text.
        none_label: str, defaults to "no_relation". The relation label that indicate dummy/negative relations.
            Predicted relations with that label will not be added to the document(s).
        max_window: int, optional. If specified, use the tokens in a window of maximal this amount of tokens
            around the center of head and tail entities and pass only that into the transformer.
        create_relation_candidates: bool, defaults to False. If True, create relation candidates by pairwise
            combining all entities in the document and assigning the none_label. If the document already contains
            a relation with the entity pair, we do not add it again. If False, assume that the document already
            contains relation annotations including negative examples (i.e. relations with the none_label).
        handle_relations_with_same_arguments: str, defaults to "keep_none". If "keep_none", all relations that
            share same arguments will be removed. If "keep_first", first occurred duplicate will be kept.
        argument_type_whitelist: List[List[str]], optional, defaults to None. If set, only relations (candidates)
            with given argument type tuples are created from document and by by `create_relation_candidates`.
            This affects only model input.
        argument_and_relation_type_whitelist: Union[Dict[str, List[List[str]]], List[List[str]]], optional,
            defaults None. If set, only given relation types with given argument types will persist in
            documents and generated by `create_relation_candidates`. This also affects predictions on
            `decode()`, so it strictly filters both model input and output. Can also be passed as a list
            of lists, where the first element is the relation type and the rest are the argument types.
    """

    PREPARED_ATTRIBUTES = ["labels", "entity_labels"]

    def __init__(
        self,
        tokenizer_name_or_path: str,
        relation_annotation: str = "binary_relations",
        add_candidate_relations: bool = False,
        add_reversed_relations: bool = False,
        partition_annotation: Optional[str] = None,
        none_label: str = "no_relation",
        padding: Union[bool, str, PaddingStrategy] = True,
        truncation: Union[bool, str, TruncationStrategy] = True,
        max_length: Optional[int] = None,
        pad_to_multiple_of: Optional[int] = None,
        multi_label: bool = False,
        labels: Optional[List[str]] = None,
        label_to_id: Optional[Dict[str, int]] = None,
        add_type_to_marker: bool = False,
        argument_role_to_marker: Optional[Dict[str, str]] = None,
        single_argument_pair: bool = True,
        append_markers: bool = False,
        insert_markers: bool = True,
        entity_labels: Optional[List[str]] = None,
        reversed_relation_label_suffix: str = "_reversed",
        symmetric_relations: Optional[List[str]] = None,
        reverse_symmetric_relations: bool = True,
        max_argument_distance: Optional[int] = None,
        max_argument_distance_type: str = "inner",
        max_argument_distance_tokens: Optional[int] = None,
        max_argument_distance_type_tokens: str = "inner",
        max_window: Optional[int] = None,
        allow_discontinuous_text: bool = False,
        log_first_n_examples: int = 0,
        add_argument_indices_to_input: bool = False,
        add_argument_tags_to_input: bool = False,
        add_entity_tags_to_input: bool = False,
        add_global_attention_mask_to_input: bool = False,
        argument_type_whitelist: Optional[List[List[str]]] = None,
        handle_relations_with_same_arguments: str = "keep_none",
        argument_and_relation_type_whitelist: Optional[
            Union[Dict[str, List[List[str]]], List[List[str]]]
        ] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if label_to_id is not None:
            logger.warning(
                "The parameter label_to_id is deprecated and will be removed in a future version. "
                "Please use labels instead."
            )
            id_to_label = {v: k for k, v in label_to_id.items()}
            # reconstruct labels from label_to_id. Note that we need to remove the none_label
            labels = [
                id_to_label[i] for i in range(len(id_to_label)) if id_to_label[i] != none_label
            ]
        self.save_hyperparameters(ignore=["label_to_id"])

        self.relation_annotation = relation_annotation
        self.add_candidate_relations = add_candidate_relations
        self.add_reversed_relations = add_reversed_relations
        self.padding = padding
        self.truncation = truncation
        self.labels = labels
        self.max_length = max_length
        self.pad_to_multiple_of = pad_to_multiple_of
        self.multi_label = multi_label
        self.add_type_to_marker = add_type_to_marker
        self.single_argument_pair = single_argument_pair
        self.append_markers = append_markers
        self.insert_markers = insert_markers
        self.entity_labels = entity_labels
        self.partition_annotation = partition_annotation
        self.none_label = none_label
        self.reversed_relation_label_suffix = reversed_relation_label_suffix
        self.symmetric_relations = set(symmetric_relations or [])
        self.reverse_symmetric_relations = reverse_symmetric_relations
        self.max_argument_distance = max_argument_distance
        self.max_argument_distance_type = max_argument_distance_type
        self.max_argument_distance_tokens = max_argument_distance_tokens
        self.max_argument_distance_type_tokens = max_argument_distance_type_tokens
        self.max_window = max_window
        self.allow_discontinuous_text = allow_discontinuous_text
        self.handle_relations_with_same_arguments = handle_relations_with_same_arguments
        self.argument_type_whitelist: Optional[Set[Tuple[str, ...]]] = None
        self.argument_and_relation_type_whitelist: Optional[Dict[str, Set[Tuple[str, ...]]]] = None

        if argument_type_whitelist is not None:
            # hydra does not support tuples, so we got lists and need to convert them
            self.argument_type_whitelist = {tuple(types) for types in argument_type_whitelist}
        if argument_and_relation_type_whitelist is not None:
            # hydra does not support tuples, so we got lists and need to convert them
            if isinstance(argument_and_relation_type_whitelist, list):
                self.argument_and_relation_type_whitelist = defaultdict(set)
                for types_list in argument_and_relation_type_whitelist:
                    if len(types_list) < 1:
                        raise ValueError(
                            "argument_and_relation_type_whitelist must be a list of lists with at least one element"
                        )
                    self.argument_and_relation_type_whitelist[types_list[0]].add(
                        tuple(types_list[1:])
                    )
            else:
                self.argument_and_relation_type_whitelist = {
                    rel: {tuple(types) for types in types_list}
                    for rel, types_list in argument_and_relation_type_whitelist.items()
                }
        # overwrite None with 0 for backward compatibility
        self.log_first_n_examples = log_first_n_examples or 0
        self.add_argument_indices_to_input = add_argument_indices_to_input
        self.add_argument_tags_to_input = add_argument_tags_to_input
        self.add_entity_tags_to_input = add_entity_tags_to_input
        self.add_global_attention_mask_to_input = add_global_attention_mask_to_input
        if argument_role_to_marker is None:
            self.argument_role_to_marker = {HEAD: "H", TAIL: "T"}
        else:
            self.argument_role_to_marker = argument_role_to_marker

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

        # used when allow_discontinuous_text
        self.glue_token_ids = self._get_glue_token_ids()

        self.argument_markers = None

        self._logged_examples_counter = 0

    def _get_glue_token_ids(self):
        dummy_ids = self.tokenizer.build_inputs_with_special_tokens(
            token_ids_0=[-1], token_ids_1=[-2]
        )
        return dummy_ids[dummy_ids.index(-1) + 1 : dummy_ids.index(-2)]

    @property
    def document_type(self) -> Optional[Type[DocumentType]]:
        if self.partition_annotation is not None:
            dt = TextDocumentWithLabeledSpansBinaryRelationsAndLabeledPartitions
        else:
            dt = TextDocumentWithLabeledSpansAndBinaryRelations
        if self.relation_annotation == "binary_relations":
            return dt
        else:
            logger.warning(
                f"relation_annotation={self.relation_annotation} is "
                f"not the default value ('binary_relations'), so the taskmodule {type(self).__name__} can not request "
                f"the usual document type for auto-conversion ({dt.__name__}) because this has the bespoken default "
                f"value as layer name instead of the provided one."
            )
            return None

    def get_relation_layer(self, document: Document) -> AnnotationLayer[BinaryRelation]:
        return document[self.relation_annotation]

    def get_entity_layer(self, document: Document) -> AnnotationLayer[LabeledSpan]:
        relations: AnnotationLayer[BinaryRelation] = self.get_relation_layer(document)
        return relations.target_layer

    def get_marker_factory(self) -> MarkerFactory:
        return MarkerFactory(role_to_marker=self.argument_role_to_marker)

    def _prepare(self, documents: Sequence[DocumentType]) -> None:
        entity_labels: Set[str] = set()
        relation_labels: Set[str] = set()
        for document in documents:
            relations: AnnotationLayer[BinaryRelation] = self.get_relation_layer(document)
            entities: AnnotationLayer[LabeledSpan] = self.get_entity_layer(document)

            for entity in entities:
                entity_labels.add(entity.label)

            for relation in relations:
                relation_labels.add(relation.label)
                if self.add_reversed_relations:
                    if relation.label.endswith(self.reversed_relation_label_suffix):
                        raise ValueError(
                            f"doc.id={document.id}: the relation label '{relation.label}' already ends with "
                            f"the reversed_relation_label_suffix '{self.reversed_relation_label_suffix}', "
                            f"this is not allowed because we would not know if we should strip the suffix and "
                            f"revert the arguments during inference or not"
                        )
                    if relation.label not in self.symmetric_relations:
                        relation_labels.add(relation.label + self.reversed_relation_label_suffix)

        if self.none_label in relation_labels:
            relation_labels.remove(self.none_label)

        self.labels = sorted(relation_labels)
        self.entity_labels = sorted(entity_labels)

    def encode(self, *args, **kwargs):
        self.reset_statistics()
        res = super().encode(*args, **kwargs)
        self.show_statistics()
        return res

    def _post_prepare(self):
        self.label_to_id = {label: i + 1 for i, label in enumerate(self.labels)}
        self.label_to_id[self.none_label] = 0
        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

        self.marker_factory = self.get_marker_factory()
        self.argument_markers = self.marker_factory.get_all_markers(
            append_markers=self.append_markers,
            add_type_to_marker=self.add_type_to_marker,
            entity_labels=self.entity_labels,
        )
        self.tokenizer.add_tokens(self.argument_markers, special_tokens=True)

        self.argument_markers_to_id = {
            marker: self.tokenizer.vocab[marker] for marker in self.argument_markers
        }

        self.argument_role2idx = {
            role: i for i, role in enumerate(sorted(self.marker_factory.all_roles))
        }

    def _add_reversed_relations(
        self,
        arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation],
        doc_id: Optional[str] = None,
    ) -> None:
        if self.add_reversed_relations:
            for arguments, rel in list(arguments2relation.items()):
                arg_roles, arg_spans = zip(*arguments)
                if isinstance(rel, BinaryRelation):
                    label = rel.label
                    if label in self.symmetric_relations and not self.reverse_symmetric_relations:
                        continue
                    if label.endswith(self.reversed_relation_label_suffix):
                        raise ValueError(
                            f"doc.id={doc_id}: The relation has the label '{label}' which already ends with the "
                            f"reversed_relation_label_suffix='{self.reversed_relation_label_suffix}'. "
                            f"It looks like the relation is already reversed, which is not allowed."
                        )
                    if rel.label not in self.symmetric_relations:
                        label += self.reversed_relation_label_suffix

                    reversed_rel = BinaryRelation(
                        head=rel.tail,
                        tail=rel.head,
                        label=label,
                        score=rel.score,
                    )
                    reversed_arguments = get_relation_argument_spans_and_roles(reversed_rel)
                    if reversed_arguments in arguments2relation:
                        prev_rel = arguments2relation[reversed_arguments]
                        prev_label = prev_rel.label
                        logger.warning(
                            f"doc.id={doc_id}: there is already a relation with reversed "
                            f"arguments={reversed_arguments} and label={prev_label}, so we do not add the reversed "
                            f"relation (with label {prev_label}) for these arguments"
                        )
                        if self.collect_statistics:
                            self.collect_relation("skipped_reversed_same_arguments", reversed_rel)
                        continue
                    elif rel.label in self.symmetric_relations:
                        # warn if the original relation arguments were not sorted by their start and end positions
                        # in the case of symmetric relations
                        if not all(isinstance(arg_span, Span) for arg_span in arg_spans):
                            raise NotImplementedError(
                                f"doc.id={doc_id}: the taskmodule does not yet support adding reversed relations "
                                f"for symmetric relations with arguments that are no Spans: {arguments}"
                            )
                        args_sorted = sorted(
                            [rel.head, rel.tail], key=lambda span: (span.start, span.end)
                        )
                        if args_sorted != [rel.head, rel.tail]:
                            logger.warning(
                                f"doc.id={doc_id}: The symmetric relation with label '{label}' has arguments "
                                f"{arguments} which are not sorted by their start and end positions. "
                                f"This may lead to problems during evaluation because we assume that the "
                                f"arguments of symmetric relations were sorted in the beginning and, thus, interpret "
                                f"relations where this is not the case as reversed. All reversed relations will get "
                                f"their arguments swapped during inference in the case of add_reversed_relations=True "
                                f"to remove duplicates. You may consider adding reversed versions of the *symmetric* "
                                f"relations on your own and then setting *reverse_symmetric_relations* to False."
                            )
                            if self.collect_statistics:
                                self.collect_relation(
                                    "used_not_sorted_reversed_arguments", reversed_rel
                                )

                    arguments2relation[reversed_arguments] = reversed_rel
                else:
                    raise NotImplementedError(
                        f"doc.id={doc_id}: the taskmodule does not yet support adding reversed relations for type: "
                        f"{type(rel)}"
                    )

    def _filter_relations_by_argument_and_relation_type_whitelist(
        self,
        arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation],
        doc_id: Optional[str] = None,
    ) -> None:
        if self.argument_and_relation_type_whitelist is not None:
            for arguments, relation in list(arguments2relation.items()):
                argument_labels = tuple(getattr(ann, "label") for role, ann in arguments)
                relation_label = getattr(relation, "label")
                if (
                    relation_label not in self.argument_and_relation_type_whitelist
                    or argument_labels
                    not in self.argument_and_relation_type_whitelist[relation_label]
                ):
                    rel = arguments2relation.pop(arguments)
                    self.collect_relation("skipped_argument_and_relation_type_whitelist", rel)

    def _filter_relations_by_argument_type_whitelist(
        self,
        arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation],
        doc_id: Optional[str] = None,
    ) -> None:
        if self.argument_type_whitelist is not None:
            for arguments, rel in list(arguments2relation.items()):
                argument_labels = tuple(getattr(arg, "label") for _, arg in arguments)
                if argument_labels not in self.argument_type_whitelist:
                    rel = arguments2relation.pop(arguments)
                    self.collect_relation("skipped_argument_type_whitelist", rel)

    def _add_candidate_relations(
        self,
        arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation],
        entities: Iterable[Span],
        arguments_blacklist: Optional[Set[Tuple[Tuple[str, Annotation], ...]]] = None,
        doc_id: Optional[str] = None,
    ) -> None:
        if self.add_candidate_relations:
            if self.marker_factory.all_roles == {HEAD, TAIL}:
                # flatten argument_and_relation_type_whitelist values
                arg_rel_whitelist_vals_set = (
                    None
                    if self.argument_and_relation_type_whitelist is None
                    else {i for j in self.argument_and_relation_type_whitelist.values() for i in j}
                )
                # iterate over all possible argument candidates
                for head in entities:
                    for tail in entities:
                        if head == tail:
                            continue

                        # Create a relation candidate with the none label. Otherwise, we use the existing relation.
                        new_relation = BinaryRelation(
                            head=head, tail=tail, label=self.none_label, score=1.0
                        )
                        new_relation_args = get_relation_argument_spans_and_roles(new_relation)
                        arg_roles, arg_spans = zip(*new_relation_args)
                        arg_labels = tuple(getattr(ann, "label") for ann in arg_spans)

                        # Skip if argument_type_whitelist and/or argument_and_relation_type_whitelist
                        # are defined and current candidates do not fit.
                        if (
                            self.argument_type_whitelist is not None
                            and arg_labels not in self.argument_type_whitelist
                        ) or (
                            arg_rel_whitelist_vals_set is not None
                            and arg_labels not in arg_rel_whitelist_vals_set
                        ):
                            continue

                        # check blacklist
                        if (
                            arguments_blacklist is not None
                            and new_relation_args in arguments_blacklist
                        ):
                            continue

                        # we use the new relation only if there is no existing relation with the same arguments
                        if new_relation_args not in arguments2relation:
                            arguments2relation[new_relation_args] = new_relation
            else:
                raise NotImplementedError(
                    f"doc.id={doc_id}: the taskmodule does not yet support adding relation candidates "
                    f"with argument roles other than 'head' and 'tail': {sorted(self.marker_factory.all_roles)}"
                )

    def _filter_relations_by_argument_distance(
        self,
        arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation],
        doc_id: Optional[str] = None,
    ) -> None:
        if self.max_argument_distance is not None:
            for arguments, rel in list(arguments2relation.items()):
                if isinstance(rel, BinaryRelation):
                    if isinstance(rel.head, Span) and isinstance(rel.tail, Span):
                        dist = span_distance(
                            (rel.head.start, rel.head.end),
                            (rel.tail.start, rel.tail.end),
                            self.max_argument_distance_type,
                        )
                        if dist > self.max_argument_distance:
                            arguments2relation.pop(arguments)
                            self.collect_relation("skipped_argument_distance", rel)
                    else:
                        raise NotImplementedError(
                            f"doc.id={doc_id}: the taskmodule does not yet support filtering relation candidates "
                            f"with arguments of type: {type(rel.head)} and {type(rel.tail)}"
                        )
                else:
                    raise NotImplementedError(
                        f"doc.id={doc_id}: the taskmodule does not yet support filtering relation candidates for "
                        f"type: {type(rel)}"
                    )

    def encode_input(
        self,
        document: DocumentType,
        is_training: bool = False,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        all_relations: Sequence[Annotation] = self.get_relation_layer(document)
        all_entities: Sequence[Span] = self.get_entity_layer(document)
        self.collect_all_relations("available", all_relations)

        partitions: Sequence[Span]
        if self.partition_annotation is not None:
            partitions = document[self.partition_annotation]
            if len(partitions) == 0:
                logger.warning(
                    f"the document {document.id} has no '{self.partition_annotation}' partition entries, "
                    f"no inputs will be created!"
                )
        else:
            # use single dummy partition
            partitions = [Span(start=0, end=len(document.text))]

        task_encodings: List[TaskEncodingType] = []
        for partition in partitions:
            # get all entities that are contained in the current partition
            entities: List[Span] = [
                entity
                for entity in all_entities
                if is_contained_in((entity.start, entity.end), (partition.start, partition.end))
            ]

            # Create a mapping from relation arguments to the respective relation objects.
            # Note that the data can contain multiple relations with the same arguments.
            entities_set = set(entities)
            arguments2relations: Dict[Tuple[Tuple[str, Annotation], ...], List[Annotation]] = (
                defaultdict(list)
            )
            for rel in all_relations:
                # Skip relations with unknown labels. Use label_to_id because that contains the none_label
                if rel.label not in self.label_to_id:
                    self.collect_relation("skipped_unknown_label", rel)
                    continue

                arguments = get_relation_argument_spans_and_roles(rel)
                arg_roles, arg_spans = zip(*arguments)

                # filter out all relations that are completely outside the current partition
                if all(arg_span not in entities_set for arg_span in arg_spans):
                    continue

                # filter relations that are only partially contained in the current partition,
                # i.e. some arguments are in the partition and some are not
                if any(arg_span not in entities_set for arg_span in arg_spans):
                    logger.warning(
                        f"doc.id={document.id}: there is a relation with label '{rel.label}' and arguments "
                        f"{arguments} that is only partially contained in the current partition. "
                        f"We skip this relation."
                    )
                    self.collect_relation("skipped_partially_contained", rel)
                    continue
                arguments2relations[arguments].append(rel)

            # resolve duplicates for same arguments
            arguments2relation: Dict[Tuple[Tuple[str, Annotation], ...], Annotation] = {}
            # we will never create an encoding for the relation candidates in arguments_blacklist
            arguments_blacklist: Set[Tuple[Tuple[str, Annotation], ...]] = set()
            for arguments, relations in arguments2relations.items():
                relations_set = set(relations)
                # more than one unique relation with the same arguments
                if len(relations_set) > 1:
                    arguments_resolved = tuple(map(lambda x: (x[0], x[1].resolve()), arguments))
                    labels = [rel.label for rel in relations]
                    if self.handle_relations_with_same_arguments == "keep_first":
                        # keep only the first relation
                        arguments2relation[arguments] = relations[0]
                        for discard_rel in set(relations) - {
                            relations[0]
                        }:  # remove all other relations
                            self.collect_relation("skipped_same_arguments", discard_rel)
                        if not self.collect_statistics:
                            # We show this warning only if statistics are disabled.
                            # We want to be informed if such skip occurs, but having it in statistics and
                            # getting lots of warnings in the same time seemed overwhelming.
                            logger.warning(
                                f"doc.id={document.id}: there are multiple relations with the same arguments "
                                f"{arguments_resolved}, but different labels: {labels}. We only keep the first "
                                f"occurring relation which has the label='{relations[0].label}'."
                            )
                    elif self.handle_relations_with_same_arguments == "keep_none":
                        # add these arguments to the blacklist to not add them as 'no-relation's back again
                        arguments_blacklist.add(arguments)
                        # remove all relations with the same arguments
                        for discard_rel in relations_set:
                            self.collect_relation("skipped_same_arguments", discard_rel)
                        if not self.collect_statistics:
                            logger.warning(
                                f"doc.id={document.id}: there are multiple relations with the same arguments "
                                f"{arguments_resolved}, but different labels: {labels}. All relations will be removed."
                            )
                    else:
                        raise ValueError(
                            f"'handle_relations_with_same_arguments' must be 'keep_first' or 'keep_none', "
                            f"but got `{self.handle_relations_with_same_arguments}`."
                        )
                else:
                    arguments2relation[arguments] = relations[0]
                    # more than one duplicate relation (with the same arguments)
                    if len(relations) > 1:
                        # if 'collect_statistics=true' such duplicates won't be collected and are not counted in
                        # statistics if 'collect_statistics=true' either as 'available' or as 'skipped_same_arguments'
                        logger.warning(
                            f"doc.id={document.id}: Relation annotation `{rel.resolve()}` is duplicated. "
                            f"We keep only one of them. Duplicate won't appear in statistics either as 'available' "
                            f"or as skipped."
                        )

            # We use this filter before adding reversed relations because we also don't want them to be reversed
            self._filter_relations_by_argument_and_relation_type_whitelist(
                arguments2relation=arguments2relation, doc_id=document.id
            )
            self._add_reversed_relations(arguments2relation=arguments2relation, doc_id=document.id)
            self._filter_relations_by_argument_type_whitelist(
                arguments2relation=arguments2relation, doc_id=document.id
            )
            self._add_candidate_relations(
                arguments2relation=arguments2relation,
                arguments_blacklist=arguments_blacklist,
                entities=entities,
                doc_id=document.id,
            )

            self._filter_relations_by_argument_distance(
                arguments2relation=arguments2relation, doc_id=document.id
            )

            without_special_tokens = self.max_window is not None
            text = document.text[partition.start : partition.end]
            encoding = self.tokenizer(
                text,
                padding=False,
                truncation=self.truncation if self.max_window is None else False,
                max_length=self.max_length,
                is_split_into_words=False,
                return_offsets_mapping=False,
                add_special_tokens=not without_special_tokens,
            )

            for arguments, rel in arguments2relation.items():
                arg_roles, arg_spans = zip(*arguments)
                if not all(isinstance(arg, LabeledSpan) for arg in arg_spans):
                    # TODO: add test case for this
                    raise ValueError(
                        f"the taskmodule expects the relation arguments to be of type LabeledSpan, "
                        f"but got {[type(arg) for arg in arg_spans]}"
                    )

                arg_spans_partition = [
                    shift_span(span, offset=-partition.start) for span in arg_spans
                ]
                # map character spans to token spans
                try:
                    arg_token_spans = [
                        get_aligned_token_span(
                            encoding=encoding,
                            char_span=arg,
                        )
                        for arg in arg_spans_partition
                    ]
                # Check if the mapping was successful. It may fail (and is None) if any argument start or end does not
                # match a token start or end, respectively.
                except SpanNotAlignedWithTokenException as e:
                    span_original = shift_span(e.span, offset=partition.start)
                    # the span is not attached because we shifted it above, so we can not use str(e.span)
                    span_text = document.text[span_original.start : span_original.end]
                    logger.warning(
                        f"doc.id={document.id}: Skipping invalid example, cannot get argument token slice for "
                        f'{span_original}: "{span_text}"'
                    )
                    self.collect_relation("skipped_args_not_aligned", rel)
                    continue

                # create the argument objects
                args = [
                    RelationArgument(
                        entity=span,
                        role=role,
                        token_span=token_span,
                        add_type_to_marker=self.add_type_to_marker,
                        marker_factory=self.marker_factory,
                    )
                    for span, role, token_span in zip(arg_spans, arg_roles, arg_token_spans)
                ]

                if self.max_argument_distance_tokens is not None:
                    token_distances = []
                    for idx1 in range(len(args) - 1):
                        for idx in range(idx1 + 1, len(args)):
                            arg1 = args[idx1]
                            arg2 = args[idx]
                            dist = span_distance(
                                (arg1.token_span.start, arg1.token_span.end),
                                (arg2.token_span.start, arg2.token_span.end),
                                self.max_argument_distance_type_tokens,
                            )
                            token_distances.append(dist)
                    if len(token_distances) > 0:
                        if self.max_argument_distance_type_tokens == "outer":
                            max_dist = max(token_distances)
                        elif self.max_argument_distance_type_tokens == "inner":
                            if len(args) > 2:
                                raise NotImplementedError(
                                    f"max_argument_distance_type_tokens={self.max_argument_distance_type_tokens} "
                                    f"is not supported for relations with more than 2 arguments"
                                )
                            max_dist = max(token_distances)
                        else:
                            raise NotImplementedError(
                                f"max_argument_distance_type_tokens={self.max_argument_distance_type_tokens} "
                                f"is not supported"
                            )
                        if max_dist > self.max_argument_distance_tokens:
                            self.collect_relation("skipped_argument_distance_tokens", rel)
                            continue

                input_ids = encoding["input_ids"]

                entity_tags = None
                if self.add_entity_tags_to_input:
                    entity_spans_partition = [
                        shift_span(span, offset=-partition.start) for span in entities
                    ]
                    entity_token_spans = []
                    for span in entity_spans_partition:
                        try:
                            entity_token_spans.append(
                                get_aligned_token_span(
                                    encoding=encoding,
                                    char_span=span,
                                )
                            )
                        except SpanNotAlignedWithTokenException as e:
                            span_original = shift_span(e.span, offset=partition.start)
                            span_text = document.text[span_original.start : span_original.end]
                            logger.warning(
                                f"doc.id={document.id}: Skipping invalid example, cannot get entity token slice for "
                                f'{span_original}: "{span_text}"'
                            )
                            self.collect_relation("skipped_entity_not_aligned", rel)
                            continue

                    entity_tags = bio_encode_spans(
                        spans=[
                            (span.start, span.end, getattr(span, "label", "ENTITY"))
                            for span in entity_token_spans
                        ],
                        total_length=len(input_ids),
                        label2idx={
                            label: idx for idx, label in enumerate(self.entity_labels or [])
                        },
                    )

                # windowing: we restrict the input to a window of a maximal size (max_window) with the arguments
                # of the candidate relation in the center (as much as possible)
                if self.max_window is not None:
                    # The actual number of tokens needs to be lower than max_window because we add two
                    # marker tokens (before / after) each argument and the default special tokens
                    # (e.g. CLS and SEP).
                    max_tokens = self.max_window - self.tokenizer.num_special_tokens_to_add()
                    if self.insert_markers:
                        max_tokens -= len(args) * 2
                    # if we add the markers also to the end, this decreases the available window again by
                    # two tokens (marker + sep) per argument
                    if self.append_markers:
                        # TODO: add test case for this
                        max_tokens -= len(args) * 2

                    if self.allow_discontinuous_text:
                        if entity_tags is not None:
                            raise NotImplementedError(
                                "allow_discontinuous_text=True is not yet supported with add_entity_tags_to_input=True"
                            )

                        max_tokens_per_argument = max_tokens // len(args)
                        max_tokens_per_argument -= len(self.glue_token_ids)
                        if any(
                            arg.token_span.end - arg.token_span.start > max_tokens_per_argument
                            for arg in args
                        ):
                            self.collect_relation("skipped_too_long_argument", rel)
                            continue

                        mask = np.zeros_like(input_ids)
                        for arg in args:
                            # if the input is already fully covered by one argument frame, we keep everything
                            if len(input_ids) <= max_tokens_per_argument:
                                mask[:] = 1
                                break
                            arg_center = (arg.token_span.end + arg.token_span.start) // 2
                            arg_frame_start = arg_center - max_tokens_per_argument // 2
                            # shift the frame to the right if it is out of bounds
                            if arg_frame_start < 0:
                                arg_frame_start = 0
                            arg_frame_end = arg_frame_start + max_tokens_per_argument
                            # shift the frame to the left if it is out of bounds
                            # Note that this can not cause to have arg_frame_start < 0 because we already
                            # checked that the frame is not larger than the input.
                            if arg_frame_end > len(input_ids):
                                arg_frame_end = len(input_ids)
                                arg_frame_start = arg_frame_end - max_tokens_per_argument
                            # still, a sanity check
                            if arg_frame_start < 0:
                                raise ValueError(
                                    f"arg_frame_start={arg_frame_start} < 0 after adjusting arg_frame_end={arg_frame_end}"
                                )
                            mask[arg_frame_start:arg_frame_end] = 1
                        offsets = np.cumsum(mask != 1)
                        arg_cluster_offset_values = set()
                        # sort by start indices
                        args_sorted = sorted(args, key=lambda x: x.token_span.start)
                        for arg in args_sorted:
                            offset = offsets[arg.token_span.start]
                            arg_cluster_offset_values.add(offset)
                            arg.shift_token_span(-offset)
                            # shift back according to inserted glue patterns
                            num_glues = len(arg_cluster_offset_values) - 1
                            arg.shift_token_span(num_glues * len(self.glue_token_ids))

                        new_input_ids: List[int] = []
                        for arg_cluster_offset_value in sorted(arg_cluster_offset_values):
                            if len(new_input_ids) > 0:
                                new_input_ids.extend(self.glue_token_ids)
                            segment_mask = offsets == arg_cluster_offset_value
                            segment_input_ids = [
                                input_id
                                for input_id, keep in zip(input_ids, mask & segment_mask)
                                if keep
                            ]
                            new_input_ids.extend(segment_input_ids)

                        input_ids = new_input_ids
                    else:
                        # the slice from the beginning of the first entity to the end of the second is required
                        slice_required = (
                            min(arg.token_span.start for arg in args),
                            max(arg.token_span.end for arg in args),
                        )
                        window_slice = get_window_around_slice(
                            slice=slice_required,
                            max_window_size=max_tokens,
                            available_input_length=len(input_ids),
                        )
                        # this happens if slice_required (all arguments) does not fit into max_tokens (the available window)
                        if window_slice is None:
                            self.collect_relation("skipped_too_long", rel)
                            continue

                        window_start, window_end = window_slice
                        input_ids = input_ids[window_start:window_end]

                        if entity_tags is not None:
                            entity_tags = entity_tags[window_start:window_end]

                        for arg in args:
                            arg.shift_token_span(-window_start)

                # collect all markers with their target positions, the source argument, and
                marker_ids_with_positions = []
                for arg in args:
                    marker_ids_with_positions.append(
                        (
                            self.argument_markers_to_id[arg.as_start_marker],
                            arg.token_span.start,
                            arg,
                            START,
                        )
                    )
                    marker_ids_with_positions.append(
                        (
                            self.argument_markers_to_id[arg.as_end_marker],
                            arg.token_span.end,
                            arg,
                            END,
                        )
                    )

                # create new input ids with the markers inserted and collect new mention offsets
                input_ids_with_markers = list(input_ids)
                offset = 0
                arg_start_indices = [-1] * len(self.argument_role2idx)
                arg_end_indices = [-1] * len(self.argument_role2idx)
                marker_ids_with_positions_sorted = sorted(
                    marker_ids_with_positions, key=lambda id_pos: id_pos[1]
                )
                for (
                    marker_id,
                    token_position,
                    arg,
                    marker_type,
                ) in marker_ids_with_positions_sorted:
                    if self.insert_markers:
                        input_ids_with_markers = (
                            input_ids_with_markers[: token_position + offset]
                            + [marker_id]
                            + input_ids_with_markers[token_position + offset :]
                        )
                        if entity_tags is not None:
                            entity_tags = (
                                entity_tags[: token_position + offset]
                                + [0]
                                + entity_tags[token_position + offset :]
                            )
                        offset += 1
                    if self.add_argument_indices_to_input or self.add_argument_tags_to_input:
                        idx = self.argument_role2idx[arg.role]
                        if marker_type == START:
                            if arg_start_indices[idx] != -1:
                                # TODO: add test case for this
                                raise ValueError(
                                    f"Trying to overwrite arg_start_indices[{idx}]={arg_start_indices[idx]} with "
                                    f"{token_position + offset} for document {document.id}"
                                )
                            arg_start_indices[idx] = token_position + offset
                        elif marker_type == END:
                            if arg_end_indices[idx] != -1:
                                # TODO: add test case for this
                                raise ValueError(
                                    f"Trying to overwrite arg_start_indices[{idx}]={arg_end_indices[idx]} with "
                                    f"{token_position + offset} for document {document.id}"
                                )
                            # -1 to undo the additional offset for the end marker which does not
                            # affect the mention offset
                            arg_end_indices[idx] = (
                                token_position + offset - (1 if self.insert_markers else 0)
                            )

                if self.append_markers:
                    if self.tokenizer.sep_token is None:
                        # TODO: add test case for this
                        raise ValueError("append_markers is True, but tokenizer has no sep_token")
                    sep_token_id = self.tokenizer.vocab[self.tokenizer.sep_token]
                    for arg in args:
                        if without_special_tokens:
                            # TODO: add test case for this
                            input_ids_with_markers.append(sep_token_id)
                            input_ids_with_markers.append(
                                self.argument_markers_to_id[arg.as_append_marker]
                            )
                        else:
                            input_ids_with_markers.append(
                                self.argument_markers_to_id[arg.as_append_marker]
                            )
                            input_ids_with_markers.append(sep_token_id)
                        if entity_tags is not None:
                            entity_tags.append(0)
                            entity_tags.append(0)

                # when windowing is used, we have to add the special tokens manually
                if without_special_tokens:
                    original_input_ids_with_markers = input_ids_with_markers
                    input_ids_with_markers = self.tokenizer.build_inputs_with_special_tokens(
                        token_ids_0=input_ids_with_markers
                    )
                    if self.add_argument_indices_to_input or self.add_argument_tags_to_input:
                        # get the number of prefix tokens
                        index_offset = find_sublist(
                            sub=original_input_ids_with_markers, bigger=input_ids_with_markers
                        )
                        if index_offset == -1:
                            raise ValueError(
                                f"Could not find the original tokens in the prefixed tokens for document {document.id}"
                            )
                        arg_start_indices = [
                            idx + index_offset if idx != -1 else -1 for idx in arg_start_indices
                        ]
                        arg_end_indices = [
                            idx + index_offset if idx != -1 else -1 for idx in arg_end_indices
                        ]
                    if entity_tags is not None:
                        special_tokens_mask = self.tokenizer.get_special_tokens_mask(
                            token_ids_0=input_ids_with_markers, already_has_special_tokens=True
                        )
                        entity_tags_with_special = self.tokenizer.build_inputs_with_special_tokens(
                            token_ids_0=entity_tags
                        )
                        entity_tags = [
                            tag if not is_special else 0
                            for tag, is_special in zip(
                                entity_tags_with_special, special_tokens_mask
                            )
                        ]

                inputs = {"input_ids": input_ids_with_markers}
                if self.add_argument_indices_to_input:
                    inputs["pooler_start_indices"] = arg_start_indices
                    inputs["pooler_end_indices"] = arg_end_indices
                if self.add_argument_tags_to_input:
                    # create bio-encoded tags for the arguments
                    # using arg_start_indices, arg_end_indices, and marker_ids_with_positions_sorted
                    argument_spans = [
                        (
                            arg_start_indices[self.argument_role2idx[arg.role]],
                            arg_end_indices[self.argument_role2idx[arg.role]],
                            arg.role,
                        )
                        for marker_id, token_position, arg, marker_type in marker_ids_with_positions_sorted
                    ]
                    argument_tag_ids = bio_encode_spans(
                        spans=argument_spans,
                        total_length=len(input_ids_with_markers),
                        label2idx=self.argument_role2idx,
                    )
                    inputs["argument_tags"] = argument_tag_ids

                if entity_tags is not None:
                    inputs["entity_tags"] = entity_tags

                task_encodings.append(
                    TaskEncoding(
                        document=document,
                        inputs=inputs,
                        metadata=({"candidate_annotation": rel}),
                    )
                )

                self.collect_relation("used", rel)

        return task_encodings

    def _maybe_log_example(
        self,
        task_encoding: TaskEncodingType,
        target: TargetEncodingType,
    ):
        """Maybe log the example."""

        # log the first n examples
        if self._logged_examples_counter < self.log_first_n_examples:
            input_ids = task_encoding.inputs["input_ids"]
            tokens = self.tokenizer.convert_ids_to_tokens(input_ids)
            target_labels = [self.id_to_label[label_id] for label_id in target]
            logger.info("*** Example ***")
            logger.info("doc id: %s", task_encoding.document.id)
            logger.info("tokens: %s", " ".join([str(x) for x in tokens]))
            logger.info("input_ids: %s", " ".join([str(x) for x in input_ids]))
            logger.info("Expected label: %s (ids = %s)", target_labels, target)

            self._logged_examples_counter += 1

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> TargetEncodingType:
        candidate_annotation = task_encoding.metadata["candidate_annotation"]
        if isinstance(candidate_annotation, (BinaryRelation, NaryRelation)):
            labels = [candidate_annotation.label]
        else:
            raise NotImplementedError(
                f"encoding the target with a candidate_annotation of another type than BinaryRelation or"
                f"NaryRelation is not yet supported. candidate_annotation has the type: "
                f"{type(candidate_annotation)}"
            )
        target = [self.label_to_id[label] for label in labels]

        self._maybe_log_example(task_encoding=task_encoding, target=target)

        return target

    def unbatch_output(self, model_output: ModelTargetType) -> Sequence[TaskOutputType]:
        unbatched_output = []
        if self.multi_label:
            raise NotImplementedError
        else:
            label_ids = model_output["labels"].detach().cpu().tolist()
            probabilities = model_output["probabilities"].detach().cpu().tolist()
            for batch_idx in range(len(label_ids)):
                label_id = label_ids[batch_idx]
                result: TaskOutputType = {
                    "labels": [self.id_to_label[label_id]],
                    "probabilities": [probabilities[batch_idx][label_id]],
                }
                unbatched_output.append(result)

        return unbatched_output

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Union[BinaryRelation, MultiLabeledBinaryRelation, NaryRelation]]]:
        candidate_annotation = task_encoding.metadata["candidate_annotation"]
        new_annotation: Union[BinaryRelation, MultiLabeledBinaryRelation, NaryRelation]
        if self.multi_label:
            raise NotImplementedError
        else:
            label = task_output["labels"][0]
            probability = (
                task_output["probabilities"][0] if "probabilities" in task_output else 1.0
            )
            if isinstance(candidate_annotation, BinaryRelation):
                head = candidate_annotation.head
                tail = candidate_annotation.tail
                # Reverse predicted reversed relations back. Serialization will remove any duplicated relations.
                if self.add_reversed_relations:
                    # TODO: add test case for this
                    if label.endswith(self.reversed_relation_label_suffix):
                        label = label[: -len(self.reversed_relation_label_suffix)]
                        head, tail = tail, head
                    # If the predicted label is symmetric, we sort the arguments by its center.
                    elif label in self.symmetric_relations and self.reverse_symmetric_relations:
                        if not (isinstance(head, Span) and isinstance(tail, Span)):
                            raise ValueError(
                                f"the taskmodule expects the relation arguments of the candidate_annotation"
                                f"to be of type Span, but got head of type: {type(head)} and tail of type: "
                                f"{type(tail)}"
                            )
                        # use a unique order for the arguments: sort by start and end positions
                        head, tail = sorted([head, tail], key=lambda span: (span.start, span.end))
                new_annotation = BinaryRelation(
                    head=head, tail=tail, label=label, score=probability
                )
            elif isinstance(candidate_annotation, NaryRelation):
                # TODO: add test case for this
                if self.add_reversed_relations:
                    raise ValueError("can not reverse a NaryRelation")
                new_annotation = NaryRelation(
                    arguments=candidate_annotation.arguments,
                    roles=candidate_annotation.roles,
                    label=label,
                    score=probability,
                )
            else:
                raise NotImplementedError(
                    f"creating a new annotation from a candidate_annotation of another type than BinaryRelation is "
                    f"not yet supported. candidate_annotation has the type: {type(candidate_annotation)}"
                )

            new_annotation_args = get_relation_argument_spans_and_roles(new_annotation)
            arg_roles, arg_spans = zip(*new_annotation_args)
            arg_labels = tuple(getattr(ann, "label") for ann in arg_spans)

            # Create annotation only if 1. and 2. are fulfilled:
            if (
                # 1. the label is not the no-relation-label,
                label != self.none_label
                #    or we did not create candidate relations,
                or not self.add_candidate_relations
            ) and (
                # 2. the argument_and_relation_type_whitelist is not set,
                self.argument_and_relation_type_whitelist is None
                #    or the label and argument types are in the whitelist
                or arg_labels in self.argument_and_relation_type_whitelist.get(label, {})
            ):
                yield self.relation_annotation, new_annotation

    def _get_global_attention(self, input_ids: torch.LongTensor) -> torch.LongTensor:
        # we want to have global attention on all marker tokens and the cls token
        positive_token_ids = list(self.argument_markers_to_id.values()) + [
            self.tokenizer.cls_token_id
        ]
        global_attention_mask = construct_mask(
            input_ids=input_ids, positive_ids=positive_token_ids
        )
        return global_attention_mask

    def collate(
        self, task_encodings: Sequence[TaskEncodingType]
    ) -> Tuple[ModelInputType, Optional[ModelTargetType]]:
        input_features = [
            {"input_ids": task_encoding.inputs["input_ids"]} for task_encoding in task_encodings
        ]

        inputs: Dict[str, torch.LongTensor] = self.tokenizer.pad(
            input_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        if self.add_argument_tags_to_input:
            argument_tags = [
                {"input_ids": task_encoding.inputs["argument_tags"]}
                for task_encoding in task_encodings
            ]
            argument_tags_padded = self.tokenizer.pad(
                argument_tags,
                padding=self.padding,
                max_length=self.max_length,
                pad_to_multiple_of=self.pad_to_multiple_of,
                return_tensors="pt",
            )
            # increase all values by 1 because 0 is used for padding
            inputs["argument_tags"] = argument_tags_padded["input_ids"] + 1
            # overwrite padding with 0
            inputs["argument_tags"][argument_tags_padded["attention_mask"] == 0] = 0

        if self.add_entity_tags_to_input:
            entity_tags = [
                {"input_ids": task_encoding.inputs["entity_tags"]}
                for task_encoding in task_encodings
            ]
            entity_tags_padded = self.tokenizer.pad(
                entity_tags,
                padding=self.padding,
                max_length=self.max_length,
                pad_to_multiple_of=self.pad_to_multiple_of,
                return_tensors="pt",
            )
            # increase all values by 1 because 0 is used for padding
            inputs["entity_tags"] = entity_tags_padded["input_ids"] + 1
            # overwrite padding with 0
            inputs["entity_tags"][entity_tags_padded["attention_mask"] == 0] = 0

        if self.add_argument_indices_to_input:
            pooler_start_indices = torch.tensor(
                [task_encoding.inputs["pooler_start_indices"] for task_encoding in task_encodings]
            )
            assert isinstance(pooler_start_indices, LongTensor)
            inputs["pooler_start_indices"] = pooler_start_indices
            pooler_end_indices = torch.tensor(
                [task_encoding.inputs["pooler_end_indices"] for task_encoding in task_encodings]
            )
            assert isinstance(pooler_end_indices, LongTensor)
            inputs["pooler_end_indices"] = pooler_end_indices

        if self.add_global_attention_mask_to_input:
            inputs["global_attention_mask"] = self._get_global_attention(
                input_ids=inputs["input_ids"]
            )

        if not task_encodings[0].has_targets:
            return inputs, None

        target_list: List[TargetEncodingType] = [
            task_encoding.targets for task_encoding in task_encodings
        ]
        targets = torch.tensor(target_list, dtype=torch.long)

        if not self.multi_label:
            targets = targets.flatten()

        assert isinstance(targets, LongTensor)
        return inputs, {"labels": targets}

    def configure_model_metric(self, stage: str) -> MetricCollection:
        if self.label_to_id is None:
            raise ValueError(
                "The taskmodule has not been prepared yet, so label_to_id is not known. "
                "Please call taskmodule.prepare(documents) before configuring the model metric "
                "or pass the labels to the taskmodule constructor an call taskmodule.post_prepare()."
            )
        # we use the length of label_to_id because that contains the none_label (in contrast to labels)
        labels = [self.id_to_label[i] for i in range(len(self.label_to_id))]
        common_metric_kwargs: dict[str, Any] = {
            "num_classes": len(labels),
            "task": "multilabel" if self.multi_label else "multiclass",
        }
        return MetricCollection(
            {
                "with_tn": WrappedMetricWithPrepareFunction(
                    metric=MetricCollection(
                        {
                            "micro/f1": F1Score(average="micro", **common_metric_kwargs),
                            "macro/f1": F1Score(average="macro", **common_metric_kwargs),
                            "f1_per_label": ClasswiseWrapper(
                                F1Score(average=None, **common_metric_kwargs),
                                labels=labels,
                                postfix="/f1",
                            ),
                        }
                    ),
                    prepare_function=_get_labels,
                ),
                # We can not easily calculate the macro f1 here, because
                # F1Score with average="macro" would still include the none_label.
                "micro/f1_without_tn": WrappedMetricWithPrepareFunction(
                    metric=F1Score(average="micro", **common_metric_kwargs),
                    prepare_together_function=partial(
                        _get_labels_together_remove_none_label,
                        none_idx=self.label_to_id[self.none_label],
                    ),
                ),
            }
        )



"""
workflow:
    Document
        -> (InputEncoding, TargetEncoding) -> TaskEncoding
            -> ModelStepInputType -> ModelBatchOutput
        -> TaskOutput
    -> Document
"""

import logging
from functools import partial
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypedDict,
    Union,
)

import torch
from pie_core import AnnotationLayer, TaskEncoding, TaskModule
from pie_core.utils.dictionary import list_of_dicts2dict_of_lists
from pie_documents.annotations import LabeledSpan
from pie_documents.document.processing import token_based_document_to_text_based
from pie_documents.documents import (
    TextBasedDocument,
    TextDocumentWithLabeledSpans,
    TextDocumentWithLabeledSpansAndLabeledPartitions,
    TokenDocumentWithLabeledSpans,
    TokenDocumentWithLabeledSpansAndLabeledPartitions,
)
from pie_documents.utils.sequence_tagging import tag_sequence_to_token_spans
from tokenizers import Encoding
from torchmetrics import F1Score, Metric, MetricCollection, Precision, Recall
from transformers import AutoTokenizer
from typing_extensions import TypeAlias

from pytorch_ie.models.simple_token_classification import InputType as ModelInputType
from pytorch_ie.models.simple_token_classification import TargetType as ModelTargetType
from pytorch_ie.taskmodules.metrics import (
    PrecisionRecallAndF1ForLabeledAnnotations,
    WrappedMetricWithPrepareFunction,
)
from pytorch_ie.utils.document import tokenize_document

DocumentType: TypeAlias = TextBasedDocument

InputEncodingType: TypeAlias = Encoding
TargetEncodingType: TypeAlias = Sequence[int]
TaskEncodingType: TypeAlias = TaskEncoding[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
]
ModelStepInputType: TypeAlias = Tuple[
    ModelInputType,
    Optional[ModelTargetType],
]
ModelOutputType: TypeAlias = ModelTargetType


class TaskOutputType(TypedDict, total=False):
    labels: torch.LongTensor
    probabilities: torch.FloatTensor


TaskModuleType: TypeAlias = TaskModule[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
    ModelStepInputType,
    ModelOutputType,
    TaskOutputType,
]

logger = logging.getLogger(__name__)


def _get_label_ids_from_model_output(
    model_output: ModelTargetType,
) -> torch.LongTensor:
    result = model_output["labels"]
    assert isinstance(result, torch.LongTensor)
    return result


def unbatch_and_decode_annotations(
    model_output: ModelOutputType,
    taskmodule: "LabeledSpanExtractionByTokenClassificationTaskModule",
) -> List[Sequence[LabeledSpan]]:
    task_outputs = taskmodule.unbatch_output(model_output)
    annotations = [
        taskmodule.decode_annotations(task_output)["labeled_spans"] for task_output in task_outputs
    ]
    return annotations


@TaskModule.register()
class LabeledSpanExtractionByTokenClassificationTaskModule(TaskModuleType):
    """Taskmodule for span prediction (e.g. NER) as token classification.

    This taskmodule expects the input documents to be of TextBasedDocument with an annotation layer of
    labeled spans (e.g. TextDocumentWithLabeledSpans). The text is tokenized using the provided tokenizer and
    the labels are converted to BIO tags.

    To handle long documents, the text can be windowed using the respective parameters for the tokenizer,
    i.e. max_length (and stride). Note, that this requires to set return_overflowing_tokens=True, otherwise just
    the first window of input tokens is considered. The windowing is done in a way that the spans are not split
    across windows. If a span is split across windows, it is ignored during training and evaluation. Thus, if you
    have long spans in your data, it is recommended to set a stride that is as large as the average span length
    to avoid missing many spans.

    If a partition annotation is provided, the taskmodule expects the input documents to be of
    TextBasedDocument with two annotation layers of labeled spans, one for the spans and one for the partitions
    (e.g. TextDocumentWithLabeledSpansAndLabeledPartitions). Then, the text is tokenized and fed to the model
    individually per partition (e.g. per sentence). This is useful for long documents that can not be processed
    by the model as a whole, but where a natural partitioning exists (e.g. sentences or paragraphs) and, thus,
    windowing is not necessary (or a combination of both can be used).

    If labels are not provided, they are collected from the data during the prepare() step. If provided, they act as
    whitelist, i.e. spans with labels that are not in the labels are ignored during training and evaluation.

    Args:
        tokenizer_name_or_path: Name or path of the HuggingFace tokenizer to use.
        span_annotation: Name of the annotation layer that contains the labeled spans. Default: "labeled_spans".
        partition_annotation: Name of the annotation layer that contains the labeled partitions. If provided, the
            text is tokenized individually per partition. Default: None.
        label_pad_id: ID of the padding tag label. The model should ignore this for training. Default: -100.
        labels: List of labels to use. If not provided, the labels are collected from the labeled span annotations
            in the data during the prepare() step. Default: None.
        include_ill_formed_predictions: Whether to include ill-formed predictions in the output. If False, the
            predictions are corrected to be well-formed. Default: True.
        tokenize_kwargs: Keyword arguments to pass to the tokenizer during tokenization. Default: None.
        pad_kwargs: Keyword arguments to pass to the tokenizer during padding. Note, that this is used to pad the
            token ids *and* the tag ids, if available (i.e. during training or evaluation). Default: None.
        combine_token_scores_method: Method to combine the token scores to a span score. Options are "mean", "max",
            "min", and "product". Default: "mean".
        log_precision_recall_metrics: Whether to log precision and recall metrics (in addition to F1) for the
            spans. Default: True.
    """

    # list of attribute names that need to be set by _prepare()
    PREPARED_ATTRIBUTES: List[str] = ["labels"]

    def __init__(
        self,
        tokenizer_name_or_path: str,
        span_annotation: str = "labeled_spans",
        partition_annotation: Optional[str] = None,
        label_pad_id: int = -100,
        labels: Optional[List[str]] = None,
        include_ill_formed_predictions: bool = True,
        tokenize_kwargs: Optional[Dict[str, Any]] = None,
        pad_kwargs: Optional[Dict[str, Any]] = None,
        combine_token_scores_method: str = "mean",
        log_precision_recall_metrics: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.span_annotation = span_annotation
        self.partition_annotation = partition_annotation
        self.labels = labels
        self.label_pad_id = label_pad_id
        self.include_ill_formed_predictions = include_ill_formed_predictions
        self.tokenize_kwargs = tokenize_kwargs or {}
        self.pad_kwargs = pad_kwargs or {}
        self.log_precision_recall_metrics = log_precision_recall_metrics
        self.combine_token_scores_method = combine_token_scores_method

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)

    @property
    def document_type(self) -> Optional[Type[TextBasedDocument]]:
        dt: Type[TextBasedDocument]
        errors = []
        if self.span_annotation != "labeled_spans":
            errors.append(
                f"span_annotation={self.span_annotation} is not the default value ('labeled_spans')"
            )
        if self.partition_annotation is None:
            dt = TextDocumentWithLabeledSpans
        else:
            if self.partition_annotation != "labeled_partitions":
                errors.append(
                    f"partition_annotation={self.partition_annotation} is not the default value "
                    f"('labeled_partitions')"
                )
            dt = TextDocumentWithLabeledSpansAndLabeledPartitions

        if len(errors) == 0:
            return dt
        else:
            logger.warning(
                f"{' and '.join(errors)}, so the taskmodule {type(self).__name__} can not request "
                f"the usual document type ({dt.__name__}) for auto-conversion because this has the bespoken default "
                f"value as layer name(s) instead of the provided one(s)."
            )
            return None

    def get_span_layer(self, document: DocumentType) -> AnnotationLayer[LabeledSpan]:
        return document[self.span_annotation]

    def _prepare(self, documents: Sequence[DocumentType]) -> None:
        # collect all possible labels
        labels: Set[str] = set()
        for document in documents:
            spans: AnnotationLayer[LabeledSpan] = self.get_span_layer(document)

            for span in spans:
                labels.add(span.label)

        self.labels = sorted(labels)
        logger.info(f"Collected {len(self.labels)} labels from the data: {self.labels}")

    def _post_prepare(self):
        # create the real token labels (BIO scheme) from the labels
        self.label_to_id = {"O": 0}
        current_id = 1
        for label in sorted(self.labels):
            for prefix in ["B", "I"]:
                self.label_to_id[f"{prefix}-{label}"] = current_id
                current_id += 1

        self.id_to_label = {v: k for k, v in self.label_to_id.items()}

    def encode_input(
        self,
        document: TextBasedDocument,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        if self.partition_annotation is None:
            tokenized_document_type = TokenDocumentWithLabeledSpans
            casted_document_type = TextDocumentWithLabeledSpans
            field_mapping = {self.span_annotation: "labeled_spans"}
        else:
            tokenized_document_type = TokenDocumentWithLabeledSpansAndLabeledPartitions
            casted_document_type = TextDocumentWithLabeledSpansAndLabeledPartitions
            field_mapping = {
                self.span_annotation: "labeled_spans",
                self.partition_annotation: "labeled_partitions",
            }
        casted_document = document.as_type(casted_document_type, field_mapping=field_mapping)
        tokenized_docs = tokenize_document(
            casted_document,
            tokenizer=self.tokenizer,
            result_document_type=tokenized_document_type,
            partition_layer=(
                "labeled_partitions" if self.partition_annotation is not None else None
            ),
            strict_span_conversion=False,
            **self.tokenize_kwargs,
        )

        task_encodings: List[TaskEncodingType] = []
        for tokenized_doc in tokenized_docs:
            task_encodings.append(
                TaskEncoding(
                    document=document,
                    inputs=tokenized_doc.metadata["tokenizer_encoding"],
                    metadata={"tokenized_document": tokenized_doc},
                )
            )

        return task_encodings

    def encode_target(
        self,
        task_encoding: TaskEncodingType,
    ) -> Optional[TargetEncodingType]:
        metadata = task_encoding.metadata
        tokenized_document = metadata["tokenized_document"]
        tokenizer_encoding: Encoding = tokenized_document.metadata["tokenizer_encoding"]

        tag_sequence = [
            None if tokenizer_encoding.special_tokens_mask[j] else "O"
            for j in range(len(tokenizer_encoding.ids))
        ]
        if self.labels is None:
            raise ValueError(
                "'labels' must be set before calling encode_target(). Was prepare() called on the taskmodule?"
            )
        sorted_spans = sorted(tokenized_document.labeled_spans, key=lambda s: (s.start, s.end))
        for span in sorted_spans:
            if span.label not in self.labels:
                continue
            start = span.start
            end = span.end
            if any(tag != "O" for tag in tag_sequence[start:end]):
                logger.warning(f"tag already assigned (current span has an overlap: {span}).")
                continue

            tag_sequence[start] = f"B-{span.label}"
            for j in range(start + 1, end):
                tag_sequence[j] = f"I-{span.label}"

        targets = [
            self.label_to_id[tag] if tag is not None else self.label_pad_id for tag in tag_sequence
        ]

        return targets

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> ModelStepInputType:
        input_encodings = [
            {
                "input_ids": task_encoding.inputs.ids,
                "attention_mask": task_encoding.inputs.attention_mask,
                "special_tokens_mask": task_encoding.inputs.special_tokens_mask,
            }
            for task_encoding in task_encodings
        ]
        inputs = self.tokenizer.pad(
            list_of_dicts2dict_of_lists(input_encodings), return_tensors="pt", **self.pad_kwargs
        )

        if not task_encodings[0].has_targets:
            return inputs, None

        tag_ids = [task_encoding.targets for task_encoding in task_encodings]
        targets = self.tokenizer.pad(
            {"input_ids": tag_ids}, return_tensors="pt", **self.pad_kwargs
        )["input_ids"]

        # set the padding label to the label_pad_token_id
        pad_mask = inputs["input_ids"] == self.tokenizer.pad_token_id
        targets[pad_mask] = self.label_pad_id

        return inputs, {"labels": targets}

    def unbatch_output(self, model_output: ModelOutputType) -> Sequence[TaskOutputType]:
        labels = model_output["labels"]
        probabilities = model_output.get("probabilities", None)
        batch_size = labels.shape[0]
        task_outputs: List[TaskOutputType] = []
        for batch_idx in range(batch_size):
            batch_labels = labels[batch_idx]
            assert isinstance(batch_labels, torch.LongTensor)
            task_output: TaskOutputType = {"labels": batch_labels}
            if probabilities is not None:
                batch_probabilities = probabilities[batch_idx]
                assert isinstance(batch_probabilities, torch.FloatTensor)
                task_output["probabilities"] = batch_probabilities
            task_outputs.append(task_output)
        return task_outputs

    def decode_annotations(self, encoding: TaskOutputType) -> Dict[str, Sequence[LabeledSpan]]:
        labels = encoding["labels"]
        tag_sequence = [
            "O" if tag_id == self.label_pad_id else self.id_to_label[tag_id]
            for tag_id in labels.tolist()
        ]
        labeled_spans: List[LabeledSpan] = []
        for label, (start, end_inclusive) in tag_sequence_to_token_spans(
            tag_sequence,
            coding_scheme="IOB2",
            include_ill_formed=self.include_ill_formed_predictions,
        ):
            end = end_inclusive + 1
            # do not set the score if the probabilities are not available
            annotation_kwargs = {}
            if encoding.get("probabilities") is not None:
                span_probabilities = encoding["probabilities"][start:end]
                span_label_ids = labels[start:end]
                # get the probabilities at the label indices
                span_label_probs = torch.stack(
                    [span_probabilities[i, l] for i, l in enumerate(span_label_ids)]
                )
                if self.combine_token_scores_method == "mean":
                    # use mean probability of the span as score
                    annotation_kwargs["score"] = span_label_probs.mean().item()
                elif self.combine_token_scores_method == "max":
                    # use max probability of the span as score
                    annotation_kwargs["score"] = span_label_probs.max().item()
                elif self.combine_token_scores_method == "min":
                    # use min probability of the span as score
                    annotation_kwargs["score"] = span_label_probs.min().item()
                elif self.combine_token_scores_method == "product":
                    # use product of probabilities of the span as score
                    annotation_kwargs["score"] = span_label_probs.prod().item()
                else:
                    raise ValueError(
                        f"combine_token_scores_method={self.combine_token_scores_method} is not supported."
                    )
            labeled_span = LabeledSpan(label=label, start=start, end=end, **annotation_kwargs)
            labeled_spans.append(labeled_span)
        return {"labeled_spans": labeled_spans}

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, LabeledSpan]]:
        tokenized_document = task_encoding.metadata["tokenized_document"]
        decoded_annotations = self.decode_annotations(task_output)

        # Note: token_based_document_to_text_based() does not yet consider predictions, so we need to clear
        # the main annotations and attach the predictions to that
        for layer_name, annotations in decoded_annotations.items():
            tokenized_document[layer_name].clear()
            for annotation in annotations:
                tokenized_document[layer_name].append(annotation)

        # we can not use self.document_type here because that may be None if self.span_annotation or
        # self.partition_annotation is not the default value
        document_type = (
            TextDocumentWithLabeledSpansAndLabeledPartitions
            if self.partition_annotation
            else TextDocumentWithLabeledSpans
        )
        untokenized_document: Union[
            TextDocumentWithLabeledSpans, TextDocumentWithLabeledSpansAndLabeledPartitions
        ] = token_based_document_to_text_based(
            tokenized_document, result_document_type=document_type
        )

        for span in untokenized_document.labeled_spans:
            # need to copy the span because it can be attached to only one document
            yield self.span_annotation, span.copy()

    def configure_model_metric(self, stage: str) -> Union[Metric, MetricCollection]:
        common_metric_kwargs: dict[str, Any] = {
            "num_classes": len(self.label_to_id),
            "task": "multiclass",
            "ignore_index": self.label_pad_id,
        }
        token_scores = MetricCollection(
            {
                "token/macro/f1": WrappedMetricWithPrepareFunction(
                    metric=F1Score(average="macro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
                "token/micro/f1": WrappedMetricWithPrepareFunction(
                    metric=F1Score(average="micro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
                "token/macro/precision": WrappedMetricWithPrepareFunction(
                    metric=Precision(average="macro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
                "token/macro/recall": WrappedMetricWithPrepareFunction(
                    metric=Recall(average="macro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
                "token/micro/precision": WrappedMetricWithPrepareFunction(
                    metric=Precision(average="micro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
                "token/micro/recall": WrappedMetricWithPrepareFunction(
                    metric=Recall(average="micro", **common_metric_kwargs),
                    prepare_function=_get_label_ids_from_model_output,
                ),
            }
        )

        span_scores = PrecisionRecallAndF1ForLabeledAnnotations(
            flatten_result_with_sep="/",
            prefix="span/",
            return_recall_and_precision=self.log_precision_recall_metrics,
        )
        span_scores_wrapped = WrappedMetricWithPrepareFunction(
            metric=span_scores,
            prepare_function=partial(unbatch_and_decode_annotations, taskmodule=self),
            prepare_does_unbatch=True,
        )
        # TODO: mypy complains that MetricCollection does not accept a list containing MetricCollection(s), but it works fine
        result = MetricCollection([token_scores, span_scores_wrapped])  # type: ignore
        return result



import logging
from typing import Any, Dict, Iterable, Optional, Sequence, Tuple

import torch
import torchmetrics
from pie_core import Model
from torch import nn
from torch.optim import AdamW
from transformers import AutoConfig, AutoModel, BatchEncoding, get_linear_schedule_with_warmup
from typing_extensions import TypeAlias

from pytorch_ie.model import PyTorchIEModel
from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses
from pytorch_ie.models.modules.mlp import MLP

ModelInputType: TypeAlias = BatchEncoding
ModelOutputType: TypeAlias = Dict[str, Any]

ModelStepInputType: TypeAlias = Tuple[
    ModelInputType,
    Optional[Sequence[Sequence[Tuple[int, int, int]]]],
]


TRAINING = "train"
VALIDATION = "val"
TEST = "test"


logger = logging.getLogger(__name__)


@Model.register()
class TransformerSpanClassificationModel(
    PyTorchIEModel, RequiresModelNameOrPath, RequiresNumClasses
):
    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        learning_rate: float = 1e-5,
        task_learning_rate: float = 1e-4,
        warmup_proportion: float = 0.1,
        ignore_index: int = 0,
        max_span_length: int = 8,
        span_length_embedding_dim: int = 150,
        t_total: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if t_total is not None:
            logger.warning(
                "t_total is deprecated, we use estimated_stepping_batches from the pytorch lightning trainer instead"
            )

        self.save_hyperparameters(ignore=["t_total"])

        self.t_total = t_total
        self.learning_rate = learning_rate
        self.task_learning_rate = task_learning_rate
        self.warmup_proportion = warmup_proportion
        self.max_span_length = max_span_length

        config = AutoConfig.from_pretrained(model_name_or_path)
        if self.is_from_pretrained:
            self.model = AutoModel.from_config(config=config)
        else:
            self.model = AutoModel.from_pretrained(model_name_or_path, config=config)

        classifier_dropout = (
            config.classifier_dropout
            if hasattr(config, "classifier_dropout") and config.classifier_dropout is not None
            else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        # TODO: properly intialize!
        # self.classifier = nn.Linear(config.hidden_size * 2 + span_length_embedding_dim, num_classes)
        self.classifier = MLP(
            input_dim=config.hidden_size * 2 + span_length_embedding_dim,
            output_dim=num_classes,
            hidden_dim=150,
            num_layers=2,
        )

        self.span_length_embedding = nn.Embedding(
            num_embeddings=max_span_length, embedding_dim=span_length_embedding_dim
        )

        self.loss_fct = nn.CrossEntropyLoss()

        self.f1 = nn.ModuleDict(
            {
                f"stage_{stage}": torchmetrics.F1Score(
                    num_classes=num_classes, ignore_index=ignore_index, task="multiclass"
                )
                for stage in [TRAINING, VALIDATION, TEST]
            }
        )

    def _start_end_and_span_length_span_index(
        self, batch_size: int, max_seq_length: int, seq_lengths: Optional[Iterable[int]] = None
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        if seq_lengths is None:
            seq_lengths = batch_size * [max_seq_length]

        start_indices = []
        end_indices = []
        span_lengths = []
        span_batch_index = []
        offsets = []
        for batch_index, seq_length in enumerate(seq_lengths):
            offset = max_seq_length * batch_index

            for span_length in range(1, self.max_span_length + 1):
                for start_index in range(seq_length + 1 - span_length):
                    end_index = start_index + span_length - 1

                    span_batch_index.append(batch_index)
                    start_indices.append(start_index)
                    end_indices.append(end_index)
                    span_lengths.append(span_length - 1)
                    offsets.append(offset)

        return (
            torch.tensor(start_indices),
            torch.tensor(end_indices),
            torch.tensor(span_lengths),
            torch.tensor(span_batch_index),
            torch.tensor(offsets),
        )

    # TODO: this should live in the taskmodule
    def _expand_target_tuples(
        self,
        target_tuples: Sequence[Sequence[Tuple[int, int, int]]],
        batch_size: int,
        max_seq_length: int,
        seq_lengths: Optional[Iterable[int]] = None,
    ) -> torch.Tensor:
        if seq_lengths is None:
            seq_lengths = batch_size * [max_seq_length]

        target = []
        for batch_index, seq_length in enumerate(seq_lengths):
            label_lookup = {
                (start, end): label for start, end, label in target_tuples[batch_index]
            }
            for span_length in range(1, self.max_span_length + 1):
                for start_index in range(seq_length + 1 - span_length):
                    end_index = start_index + span_length - 1

                    label = label_lookup.get((start_index, end_index), 0)
                    target.append(label)

        return torch.tensor(target)

    def forward(self, inputs: ModelInputType) -> ModelOutputType:
        output = self.model(**inputs)

        batch_size, seq_length, hidden_dim = output.last_hidden_state.shape
        hidden_state = output.last_hidden_state.view(batch_size * seq_length, hidden_dim)

        seq_lengths = None
        if "attention_mask" in inputs:
            seq_lengths = torch.sum(inputs["attention_mask"], dim=-1).detach().cpu()

        (
            start_indices,
            end_indices,
            span_length,
            batch_indices,
            offsets,
        ) = self._start_end_and_span_length_span_index(
            batch_size=batch_size, max_seq_length=seq_length, seq_lengths=seq_lengths
        )

        start_embedding = hidden_state[offsets + start_indices, :]
        end_embedding = hidden_state[offsets + end_indices, :]
        span_length_embedding = self.span_length_embedding(span_length.to(hidden_state.device))

        combined_embedding = torch.cat(
            (start_embedding, end_embedding, span_length_embedding), dim=-1
        )

        logits = self.classifier(self.dropout(combined_embedding))

        return {
            "logits": logits,
            "batch_indices": batch_indices,
            "start_indices": start_indices,
            "end_indices": end_indices,
        }

    def step(self, stage: str, batch: ModelStepInputType, batch_idx):
        inputs, target_tuples = batch
        assert target_tuples is not None, f"target has to be available for {stage}"

        output = self(inputs)

        logits = output["logits"]

        batch_size, seq_length = inputs["input_ids"].shape
        seq_lengths = None
        if "attention_mask" in inputs:
            seq_lengths = torch.sum(inputs["attention_mask"], dim=-1)

        # TODO: Why is this not happening in TransformerSpanClassificationTaskModule.collate?
        target = self._expand_target_tuples(
            target_tuples=target_tuples,
            batch_size=batch_size,
            max_seq_length=seq_length,
            seq_lengths=seq_lengths,
        )
        target = target.to(logits.device)

        loss = self.loss_fct(logits, target)

        self.log(f"{stage}/loss", loss, on_step=stage == TRAINING, on_epoch=True, prog_bar=True)

        f1 = self.f1[f"stage_{stage}"]
        f1(logits, target)
        self.log(f"{stage}/f1", f1, on_step=False, on_epoch=True, prog_bar=True)

        return loss

    def training_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TRAINING, batch=batch, batch_idx=batch_idx)

    def validation_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=VALIDATION, batch=batch, batch_idx=batch_idx)

    def test_step(self, batch: ModelStepInputType, batch_idx: int):
        return self.step(stage=TEST, batch=batch, batch_idx=batch_idx)

    def configure_optimizers(self):
        param_optimizer = list(self.named_parameters())
        # TODO: this needs fixing (does not work models other than BERT)
        optimizer_grouped_parameters = [
            {"params": [p for n, p in param_optimizer if "bert" in n]},
            {
                "params": [p for n, p in param_optimizer if "bert" not in n],
                "lr": self.task_learning_rate,
            },
        ]
        optimizer = AdamW(optimizer_grouped_parameters, lr=self.learning_rate)
        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



from .precision_recall_and_f1_for_labeled_annotations import (
    PrecisionRecallAndF1ForLabeledAnnotations,
)
from .wrapped_layer_metrics_with_unbatch_and_decode_with_errors_function import (
    WrappedLayerMetricsWithUnbatchAndDecodeWithErrorsFunction,
)
from .wrapped_metric_with_prepare_function import WrappedMetricWithPrepareFunction



from abc import ABC, abstractmethod
from typing import Generic, Iterable, Iterator, Optional, Sequence, Type, TypeVar, Union

from pie_core import Document, TaskEncoding, TaskEncodingSequence, TaskModule
from typing_extensions import TypeAlias

DocumentType = TypeVar("DocumentType", bound=Document)
ConvertedDocumentType = TypeVar("ConvertedDocumentType", bound=Document)
InputEncodingType = TypeVar("InputEncodingType")
TargetEncodingType = TypeVar("TargetEncodingType")
# TaskEncoding: defined below
TaskBatchEncodingType = TypeVar("TaskBatchEncodingType")
# ModelBatchEncoding: defined in models
ModelBatchOutputType = TypeVar("ModelBatchOutputType")
TaskOutputType = TypeVar("TaskOutputType")

TaskEncodingType: TypeAlias = TaskEncoding[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
]


class TaskModuleWithDocumentConverter(
    TaskModule,
    ABC,
    Generic[
        ConvertedDocumentType,
        DocumentType,
        InputEncodingType,
        TargetEncodingType,
        TaskBatchEncodingType,
        ModelBatchOutputType,
        TaskOutputType,
    ],
):
    @property
    def document_type(self) -> Optional[Type[Document]]:
        if super().document_type is not None:
            raise NotImplementedError(f"please overwrite document_type for {type(self).__name__}")
        else:
            return None

    @abstractmethod
    def _convert_document(self, document: DocumentType) -> ConvertedDocumentType:
        """Convert a document of the taskmodule document type to the expected document type of the
        wrapped taskmodule.

        Args:
            document: the input document

        Returns: the converted document
        """
        pass

    def _prepare(self, documents: Sequence[DocumentType]) -> None:
        # use an iterator for lazy processing
        documents_converted = (self._convert_document(doc) for doc in documents)
        super()._prepare(documents=documents_converted)

    def convert_document(self, document: DocumentType) -> ConvertedDocumentType:
        converted_document = self._convert_document(document)
        if "original_document" in converted_document.metadata:
            raise ValueError(
                f"metadata of converted_document has already and entry 'original_document', "
                f"this is not allowed. Please adjust '{type(self).__name__}._convert_document()' "
                f"to produce documents without that key in metadata."
            )
        converted_document.metadata["original_document"] = document
        return converted_document

    def encode(self, documents: Union[DocumentType, Iterable[DocumentType]], **kwargs) -> Union[
        Sequence[TaskEncodingType],
        TaskEncodingSequence[TaskEncodingType, DocumentType],
        Iterator[TaskEncodingType],
    ]:
        converted_documents: Union[DocumentType, Iterable[DocumentType]]
        if isinstance(documents, Document):
            converted_documents = self.convert_document(documents)
        else:
            converted_documents = [self.convert_document(doc) for doc in documents]
        return super().encode(documents=converted_documents, **kwargs)

    def decode(self, **kwargs) -> Sequence[DocumentType]:
        decoded_documents = super().decode(**kwargs)
        result = []
        for doc in decoded_documents:
            original_document = doc.metadata["original_document"]
            self._integrate_predictions_from_converted_document(
                converted_document=doc, document=original_document
            )
            result.append(original_document)
        return result

    @abstractmethod
    def _integrate_predictions_from_converted_document(
        self,
        document: DocumentType,
        converted_document: ConvertedDocumentType,
    ) -> None:
        """Convert the predictions at the respective layers of the converted_document and add them
        to the original document predictions.

        Args:
            document: document to attach the converted predictions to
            converted_document: the document returned by the wrapped taskmodule, including predictions
        """
        pass



# flake8: noqa

from pytorch_ie.core import *
from pytorch_ie.datamodule import PieDataModule
from pytorch_ie.dataset import IterableTaskEncodingDataset, TaskEncodingDataset
from pytorch_ie.model import PyTorchIEModel
from pytorch_ie.pipeline import PyTorchIEPipeline



import dataclasses
import logging
from functools import partial
from typing import Any, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Type, Union

import torch
from pie_core import Annotation, AnnotationLayer, Document, TaskEncoding, TaskModule
from pie_core.taskmodule import InputEncoding, ModelBatchOutput, TargetEncoding, TaskBatchEncoding
from pie_core.utils.hydra import resolve_type
from pie_documents.annotations import AnnotationWithText
from pie_documents.document.processing import token_based_document_to_text_based
from pie_documents.documents import TextBasedDocument, TokenBasedDocument
from torchmetrics import Metric
from transformers import AutoTokenizer, PreTrainedTokenizer
from typing_extensions import TypeAlias

from ..utils.document import tokenize_document
from .common import BatchableMixin, get_first_occurrence_index
from .metrics import WrappedMetricWithPrepareFunction

logger = logging.getLogger(__name__)


DocumentType: TypeAlias = TextBasedDocument


@dataclasses.dataclass
class InputEncodingType(BatchableMixin):
    input_ids: List[int]
    attention_mask: List[int]


@dataclasses.dataclass
class TargetEncodingType(BatchableMixin):
    labels: List[int]
    # this is optional because we use the same type for TaskOutputType, which does not have this field
    decoder_attention_mask: Optional[List[int]] = None


TaskEncodingType: TypeAlias = TaskEncoding[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
]
TaskOutputType: TypeAlias = TargetEncodingType


# we use a custom un-batch function for metrics, because the text metrics such as ROUGEScore metric expects
# strings for input and target
def unbatch_and_untokenize(
    batch: ModelBatchOutput, taskmodule: "TextToTextTaskModule"
) -> Sequence[str]:
    unbatched = taskmodule.unbatch_output(batch)
    texts = [
        taskmodule.tokenizer.decode(encoding.labels, skip_special_tokens=True)
        for encoding in unbatched
    ]
    return texts


@TaskModule.register()
class TextToTextTaskModule(
    TaskModule[
        DocumentType,
        InputEncoding,
        TargetEncoding,
        TaskBatchEncoding,
        ModelBatchOutput,
        TaskOutputType,
    ],
):
    """A PIE task module for text-to-text tasks. It works with simple text annotations, e.g.
    abstractive summaries, as target annotations.

    It can also be used with additional guidance annotations, e.g. questions for generative question answering, in
    which case the text of the guidance annotation is prepended to the input text.

    Args:
        tokenizer_name_or_path: The name (Huggingface Hub model identifier) or local path of the tokenizer to use.
        document_type: The type of the input document. Must be a string that resolves to a subclass of
            TextBasedDocument, e.g. "pie_documents.documents.TextDocumentWithAbstractiveSummary" for abstractive
            summarization.
        tokenized_document_type: The type of the tokenized document. Must be a string that resolves to a
            subclass of TokenBasedDocument, e.g. "pie_documents.documents.TokenDocumentWithAbstractiveSummary" for
            abstractive summarization.
        target_layer: The name of the annotation layer that contains the target annotations, e.g. "abstractive_summary"
            for abstractive summarization.
        target_annotation_type: The type of the target annotations. Must be a string that resolves to a subclass
            of AnnotationWithText, e.g. "pie_documents.annotations.AbstractiveSummary" for abstractive summarization.
        guidance_layer: The name of the annotation layer that contains the guidance annotations. If set, the text of
            the guidance annotation is prepended to the input text.
        guidance_annotation_field: The name of the field in the target annotations that contains the guidance
            annotation. Required if guidance_layer is defined to attach the guidance annotation to the newly created
            target annotation.
        text_metric_type: The type of the text metric to use for evaluation. Must be a string that resolves to a
            subclass of Metric, e.g. "torchmetrics.text.ROUGEScore" for ROUGE score.
        tokenizer_init_kwargs: Additional keyword arguments that are passed to the tokenizer constructor.
        tokenizer_kwargs: Additional keyword arguments that are passed when calling the tokenizer.
        partition_layer_name: The name of the annotation layer that contains the partitions. If set, the partitions
            will be used to split the input text into multiple parts which are then tokenized separately. This can be
            used to split long documents into multiple parts to avoid exceeding the maximum input length of the
            tokenizer / model.
        annotation_field_mapping: A mapping from input document annotation layer names to layer names defined in the
            document_type / tokenized_document_type. This can be used if the actual input documents have different
            annotation layer names than the provided document_type / tokenized_document_type.
        log_first_n_examples: The number of examples to log. If set to a positive integer n, the first n examples will
            be logged. This can be used to check if the input and target encodings are as expected.
    """

    def __init__(
        self,
        tokenizer_name_or_path: str,
        document_type: str,
        tokenized_document_type: str,
        target_layer: str,
        target_annotation_type: str,
        guidance_layer: Optional[str] = None,
        guidance_annotation_field: Optional[str] = None,
        text_metric_type: Optional[str] = None,
        tokenizer_init_kwargs: Optional[Dict[str, Any]] = None,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        partition_layer_name: Optional[str] = None,
        annotation_field_mapping: Optional[Dict[str, str]] = None,
        log_first_n_examples: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.target_layer = target_layer
        self.guidance_layer = guidance_layer
        self.target_annotation_type: Type[AnnotationWithText] = resolve_type(
            target_annotation_type, expected_super_type=AnnotationWithText
        )
        self.guidance_annotation_field = guidance_annotation_field
        self.text_metric_type: Optional[Metric] = None
        if text_metric_type is not None:
            self.text_metric_type = resolve_type(text_metric_type, expected_super_type=Metric)

        # tokenization
        self._document_type: Type[TextBasedDocument] = resolve_type(
            document_type, expected_super_type=TextBasedDocument
        )
        self._tokenized_document_type: Type[TokenBasedDocument] = resolve_type(
            tokenized_document_type, expected_super_type=TokenBasedDocument
        )
        self.tokenizer_name_or_path = tokenizer_name_or_path
        self.tokenizer_kwargs = tokenizer_kwargs or {}
        self.tokenizer: PreTrainedTokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name_or_path,
            **(tokenizer_init_kwargs or {}),
        )
        self.annotation_field_mapping = annotation_field_mapping or dict()
        self.partition_layer_name = partition_layer_name

        # target encoding
        self.pad_values = {
            "input_ids": self.tokenizer.pad_token_id,
            "attention_mask": 0,
            "labels": self.tokenizer.pad_token_id,
            "decoder_attention_mask": 0,
        }
        self.dtypes = {
            "input_ids": torch.int64,
            "attention_mask": torch.int64,
            "labels": torch.int64,
            "decoder_attention_mask": torch.int64,
        }

        # logging
        self.log_first_n_examples = log_first_n_examples

    @property
    def document_type(self) -> Type[TextBasedDocument]:
        return self._document_type

    @property
    def tokenized_document_type(self) -> Type[TokenBasedDocument]:
        return self._tokenized_document_type

    @property
    def layer_names(self) -> List[str]:
        return [self.target_layer]

    def get_mapped_layer(self, document: Document, layer_name: str) -> AnnotationLayer:
        if layer_name in self.annotation_field_mapping:
            layer_name = self.annotation_field_mapping[layer_name]
        return document[layer_name]

    @property
    def generation_config(self) -> Dict[str, Any]:
        return {}

    def maybe_log_example(
        self,
        task_encoding: TaskEncodingType,
        targets: Optional[TargetEncodingType] = None,
    ) -> None:
        if self.log_first_n_examples is not None and self.log_first_n_examples > 0:
            inputs = task_encoding.inputs

            logger.info(f"input_ids: {inputs.input_ids}")
            logger.info(f"attention_mask: {inputs.attention_mask}")
            if targets is not None or task_encoding.has_targets:
                targets = targets or task_encoding.targets
                logger.info(f"labels: {targets.labels}")
            self.log_first_n_examples -= 1

    def warn_only_once(self, message: str) -> None:
        if not hasattr(self, "_warned"):
            self._warned: Set[str] = set()
        if message not in self._warned:
            logger.warning(f"{message} (This warning will only be shown once)")
            self._warned.add(message)

    def encode_annotations(
        self,
        layers: Dict[str, AnnotationLayer],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> TargetEncodingType:
        target_annotations = []
        guidance_annotation = (
            metadata.get("guidance_annotation", None) if metadata is not None else None
        )
        if guidance_annotation is not None:
            if self.guidance_annotation_field is None:
                raise ValueError(
                    "guidance_annotation is available, but guidance_annotation_field is not set"
                )
            # filter annotations that belong to the guidance_annotation
            for target_annotation in layers[self.target_layer]:
                current_guidance_annotation = getattr(
                    target_annotation, self.guidance_annotation_field
                )
                if current_guidance_annotation == guidance_annotation:
                    target_annotations.append(target_annotation)
        else:
            target_annotations = layers[self.target_layer]

        if len(target_annotations) == 0:
            raise ValueError(f"target_annotations {self.target_layer} contains no annotation")
        elif len(target_annotations) > 1:
            self.warn_only_once(
                f"target_annotations {self.target_layer} contains more than one annotation, "
                f"but only the first one will be used"
            )
        annotation = target_annotations[0]
        if isinstance(annotation, self.target_annotation_type):
            text = target_annotations[0].text
        else:
            raise ValueError(
                f"target_annotations {self.target_layer} contains an annotation of type {type(annotation)}, "
                f"but expected {self.target_annotation_type}"
            )
        encoding = self.tokenizer(text)
        return TargetEncodingType(
            labels=encoding["input_ids"], decoder_attention_mask=encoding["attention_mask"]
        )

    def decode_annotations(
        self, encoding: TaskOutputType, metadata: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, List[Annotation]], Any]:
        text = self.tokenizer.decode(encoding.labels, skip_special_tokens=True)
        annotation_kwargs = {}
        if self.guidance_annotation_field is not None:
            if metadata is None:
                raise ValueError(
                    "metadata is required to decode annotations with guidance_annotation_field"
                )
            guidance_annotation = metadata.get("guidance_annotation", None)
            if guidance_annotation is not None:
                if self.guidance_annotation_field is None:
                    raise ValueError(
                        "guidance_annotation is available, but guidance_annotation_field is not set"
                    )
                annotation_kwargs[self.guidance_annotation_field] = guidance_annotation

        decoded_layers = {
            self.target_layer: [self.target_annotation_type(text=text, **annotation_kwargs)]
        }
        # no error collection yet
        errors: Dict[str, Any] = {}
        return decoded_layers, errors

    def tokenize_document(
        self, document: DocumentType, source_text: Optional[str] = None
    ) -> List[TokenBasedDocument]:
        field_mapping = dict(self.annotation_field_mapping)
        if self.partition_layer_name is not None:
            field_mapping[self.partition_layer_name] = "labeled_partitions"
            partition_layer = "labeled_partitions"
        else:
            partition_layer = None
        casted_document = document.as_type(self.document_type, field_mapping=field_mapping)

        tokenizer_kwargs = dict(self.tokenizer_kwargs)
        if source_text is not None:
            tokenizer_kwargs["text"] = source_text
        tokenized_docs = tokenize_document(
            casted_document,
            tokenizer=self.tokenizer,
            result_document_type=self.tokenized_document_type,
            partition_layer=partition_layer,
            **tokenizer_kwargs,
        )
        for idx, tokenized_doc in enumerate(tokenized_docs):
            tokenized_doc.id = f"{document.id}-tokenized-{idx+1}-of-{len(tokenized_docs)}"

        return tokenized_docs

    def encode_input(
        self, document: DocumentType, is_training: bool = False
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        task_encodings: List[TaskEncodingType] = []
        if self.guidance_layer is None:
            guidance_annotations = [None]
        else:
            guidance_annotations = document[self.guidance_layer]
        for guidance_annotation in guidance_annotations:
            source_text = None
            if guidance_annotation is not None:
                # Here could also more sophisticated logic be implemented
                source_text = guidance_annotation.text
            tokenized_docs = self.tokenize_document(document, source_text=source_text)
            for tokenized_doc in tokenized_docs:
                tokenizer_encoding = tokenized_doc.metadata["tokenizer_encoding"]
                task_encodings.append(
                    TaskEncoding(
                        document=document,
                        inputs=InputEncodingType(
                            input_ids=tokenizer_encoding.ids,
                            attention_mask=tokenizer_encoding.attention_mask,
                        ),
                        metadata={
                            "tokenized_document": tokenized_doc,
                            "guidance_annotation": guidance_annotation,
                        },
                    )
                )

        return task_encodings

    def encode_target(self, task_encoding: TaskEncodingType) -> Optional[TargetEncodingType]:
        document = task_encoding.metadata["tokenized_document"]
        guidance_annotation = task_encoding.metadata["guidance_annotation"]

        layers = {
            layer_name: self.get_mapped_layer(document, layer_name=layer_name)
            for layer_name in self.layer_names
        }
        result = self.encode_annotations(
            layers=layers,
            metadata={**task_encoding.metadata, "guidance_annotation": guidance_annotation},
        )

        self.maybe_log_example(task_encoding=task_encoding, targets=result)
        return result

    def collate(self, task_encodings: Sequence[TaskEncodingType]) -> TaskBatchEncoding:
        if len(task_encodings) == 0:
            raise ValueError("no task_encodings available")
        inputs = InputEncodingType.batch(
            values=[x.inputs for x in task_encodings],
            dtypes=self.dtypes,
            pad_values=self.pad_values,
        )

        targets = None
        if task_encodings[0].has_targets:
            targets = TargetEncodingType.batch(
                values=[x.targets for x in task_encodings],
                dtypes=self.dtypes,
                pad_values=self.pad_values,
            )

        return inputs, targets

    def unbatch_output(self, model_output: ModelBatchOutput) -> Sequence[TaskOutputType]:
        labels = model_output["labels"]
        batch_size = labels.size(0)

        # We use the position after the first eos token as the seq_len.
        # Note that, if eos_id is not in model_output for a given batch item, the result will be
        # model_output.size(1) + 1 (i.e. seq_len + 1) for that batch item. This is fine, because we use the
        # seq_lengths just to truncate the output and want to keep everything if eos_id is not present.
        seq_lengths = get_first_occurrence_index(labels, self.tokenizer.eos_token_id) + 1

        result = [
            TaskOutputType(labels[i, : seq_lengths[i]].to(device="cpu").tolist())
            for i in range(batch_size)
        ]
        return result

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncodingType,
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Annotation]]:
        layers, errors = self.decode_annotations(
            encoding=task_output, metadata=task_encoding.metadata
        )
        tokenized_document = task_encoding.metadata["tokenized_document"]

        # Note: token_based_document_to_text_based() does not yet consider predictions, so we need to clear
        # the main annotations and attach the predictions to that
        for layer_name, annotations in layers.items():
            layer = self.get_mapped_layer(tokenized_document, layer_name=layer_name)
            layer.clear()
            layer.extend(annotations)

        untokenized_document = token_based_document_to_text_based(
            tokenized_document, result_document_type=self.document_type
        )

        for layer_name in layers:
            annotations = self.get_mapped_layer(untokenized_document, layer_name=layer_name)
            for annotation in annotations:
                yield layer_name, annotation.copy()

    def configure_model_generation(self) -> Optional[Dict[str, Any]]:
        # we do not set any overrides here, because we want to use the default generation config as
        # it is derived from the Huggingface base model config.json
        return {}

    def configure_model_metric(self, stage: str) -> Optional[Metric]:
        if self.text_metric_type is None:
            return None

        return WrappedMetricWithPrepareFunction(
            metric=self.text_metric_type(),
            prepare_function=partial(unbatch_and_untokenize, taskmodule=self),
            prepare_does_unbatch=True,
        )



import functools
import json
import logging
from collections import defaultdict
from copy import copy
from typing import Callable, Dict, Hashable, Iterable, List, Optional, Type, TypeVar

from pie_core import Annotation, Document
from pie_core.document import BaseAnnotationList
from pie_documents.annotations import Span
from pie_documents.document.processing import text_based_document_to_token_based
from pie_documents.documents import TextBasedDocument, TokenBasedDocument, WithMetadata
from transformers import BatchEncoding, PreTrainedTokenizer

logger = logging.getLogger(__name__)


def deduplicate_annotation_dicts(
    annotation_dicts: List[Dict[str, Hashable]]
) -> List[Dict[str, Hashable]]:
    """Remove duplicate annotation dictionaries from a list of annotation dictionaries.

    Args:
        annotation_dicts: The list of annotation dictionaries to remove duplicates from.

    Returns:
        The list of annotation dictionaries with duplicates removed.
    """
    unique_annotation_dicts = []
    seen_annotation_dicts = set()
    for annotation_dict in annotation_dicts:
        annotation_dict_tuple = tuple(sorted(annotation_dict.items()))
        if annotation_dict_tuple not in seen_annotation_dicts:
            unique_annotation_dicts.append(annotation_dict)
            seen_annotation_dicts.add(annotation_dict_tuple)
    return unique_annotation_dicts


D = TypeVar("D", bound=Document)


def save_annotation_sources_to_metadata(
    document: D,
    annotation_id2source: Dict[int, List[str]],
    metadata_key: str,
    use_predictions: bool,
) -> None:
    """Save the source names for the annotations or predictions in the metadata of the document.

    Args:
        document: The document to save the source names in the metadata for.
        metadata_key: The key in the metadata where the source names should be stored.
        annotation_id2source: A mapping from annotation IDs to the source names. Should contain
            the ids of all annotations or predictions (depending on use_predictions) in the document.
        use_predictions: Whether to store the source names for the predictions or the annotations.
    """

    if not hasattr(document, "metadata"):
        raise ValueError("Document does not have metadata, can not store source names.")
    if metadata_key in document.metadata:
        raise ValueError(f"Metadata key '{metadata_key}' already exists in the document.")
    document.metadata[metadata_key] = defaultdict(dict)
    for annotation_field in document.annotation_fields():
        layer_name = annotation_field.name
        document.metadata[metadata_key][layer_name] = []
        layer: BaseAnnotationList
        if use_predictions:
            layer = document[layer_name].predictions
        else:
            layer = document[layer_name]
        for ann in layer:
            document.metadata[metadata_key][layer_name].append(annotation_id2source[ann._id])
    document.metadata[metadata_key] = dict(document.metadata[metadata_key])


def merge_annotations_from_documents(
    documents: Dict[str, D],
    metadata_key_source_annotations: Optional[str] = None,
    metadata_key_source_predictions: Optional[str] = None,
) -> D:
    """Merge annotations from multiple documents into a single document. Optionally, store the source
    names for all annotations / predictions in the metadata at key metadata_key_source_annotations
    / metadata_key_source_predictions, respectively.

    Note that this will remove any annotation duplicates.

    Args:
        documents: A dictionary mapping document source (e.g. dataset names) to documents.
        metadata_key_source_annotations: If not None, the key in the metadata where the source names
            for the (gold) annotations are stored.
        metadata_key_source_predictions: If not None, the key in the metadata where the source names
            for the predictions are stored.

    Returns:
        The document with merged annotations.
    """
    if len(documents) == 0:
        raise ValueError("No documents provided.")
    source_names = sorted(documents)
    first_source_name = source_names[0]
    merged_document: D = documents[first_source_name].copy(with_annotations=False)

    added_annotation_id2source_names: Dict[int, List[str]] = defaultdict(list)
    for source_name in source_names:
        document = documents[source_name]
        if type(document) is not type(merged_document):
            raise ValueError(
                f"Document types do not match: {type(document)} and {type(merged_document)}"
            )
        if isinstance(document, WithMetadata) and document.id is not None:
            if document.id != merged_document.id:
                raise ValueError(
                    f"Document IDs do not match: {document.id} and {merged_document.id}"
                )

        # Note: this does not check for duplicates!
        added_annotations = merged_document.add_all_annotations_from_other(
            other=document, strict=True
        )

        for layer_name, orig_id2new_annotation in added_annotations.items():
            for orig_id, new_annotation in orig_id2new_annotation.items():
                added_annotation_id2source_names[new_annotation._id].append(source_name)

    # this will remove duplicates. If duplicates have different scores, the one with the highest score will be kept
    merged_document = merged_document.deduplicate_annotations()

    # save source names in metadata (at key metadata_key_source_annotations / metadata_key_source_predictions
    #   for each layer in the order of the annotations / predictions)
    if metadata_key_source_annotations is not None:
        save_annotation_sources_to_metadata(
            document=merged_document,
            annotation_id2source=added_annotation_id2source_names,
            metadata_key=metadata_key_source_annotations,
            use_predictions=False,
        )
    if metadata_key_source_predictions is not None:
        save_annotation_sources_to_metadata(
            document=merged_document,
            annotation_id2source=added_annotation_id2source_names,
            metadata_key=metadata_key_source_predictions,
            use_predictions=True,
        )
    return merged_document


ToD = TypeVar("ToD", bound=TokenBasedDocument)


def tokenize_document(
    doc: TextBasedDocument,
    tokenizer: PreTrainedTokenizer,
    result_document_type: Type[ToD],
    partition_layer: Optional[str] = None,
    strip_spans: bool = False,
    strict_span_conversion: bool = True,
    added_annotations: Optional[List[Dict[str, Dict[Annotation, Annotation]]]] = None,
    verbose: bool = True,
    **tokenize_kwargs,
) -> List[ToD]:
    """Tokenize a document with a given tokenizer and return a list of token based documents. The
    document is tokenized in partitions if a partition layer is provided. The annotations that
    target the text are converted to target the tokens and also all dependent annotations are
    converted.

    Args:
        doc (TextBasedDocument): The document to tokenize.
        tokenizer (PreTrainedTokenizer): The tokenizer.
        result_document_type (Type[ToD]): The exact type of the token based documents.
        partition_layer (Optional[str], optional): The layer to use for partitioning the document. If None, the whole
            document is tokenized. Defaults to None.
        strip_spans (bool, optional): If True, strip the whitespace from the character spans before converting them to
            token spans. Defaults to False.
        strict_span_conversion (bool, optional): If True, raise an error if not all annotations can be converted to
            token based documents. Defaults to True.
        added_annotations (Optional[List[Dict[str, Dict[Annotation, Annotation]]]], optional): Pass an empty list to
            collect the added annotations. Defaults to None.
        verbose (bool, optional): If True, log warnings if annotations can not be converted. Defaults to True.

    Returns:
        List[ToD]: The token based documents of type result_document_type with the converted annotations.
    """

    added_annotation_lists: Dict[str, List[Annotation]] = defaultdict(list)
    result = []
    partitions: Iterable[Span]
    if partition_layer is None:
        partitions = [Span(start=0, end=len(doc.text))]
    else:
        partitions = doc[partition_layer]
    for partition in partitions:
        text = doc.text[partition.start : partition.end]
        current_tokenize_kwargs = copy(tokenize_kwargs)
        if "text" in tokenize_kwargs:
            current_tokenize_kwargs["text_pair"] = text
            sequence_index = 1
        else:
            current_tokenize_kwargs["text"] = text
            sequence_index = 0
        tokenized_text = tokenizer(**current_tokenize_kwargs)
        for batch_encoding in tokenized_text.encodings:
            token_offset_mapping = batch_encoding.offsets
            char_to_token: Optional[Callable[[int], Optional[int]]]
            char_to_token = functools.partial(
                batch_encoding.char_to_token, sequence_index=sequence_index
            )
            token_offset_mapping = [
                offsets if s_id == sequence_index else (0, 0)
                for s_id, offsets in zip(batch_encoding.sequence_ids, token_offset_mapping)
            ]
            if partition.start > 0:
                token_offset_mapping = [
                    (start + partition.start, end + partition.start)
                    for start, end in token_offset_mapping
                ]
                char_to_token = None
            current_added_annotations: Dict[str, Dict[Annotation, Annotation]] = defaultdict(dict)
            tokenized_document = text_based_document_to_token_based(
                doc,
                tokens=batch_encoding.tokens,
                result_document_type=result_document_type,
                token_offset_mapping=token_offset_mapping,
                char_to_token=char_to_token,
                strict_span_conversion=False,
                strip_spans=strip_spans,
                verbose=False,
                added_annotations=current_added_annotations,
            )
            tokenized_document.metadata["tokenizer_encoding"] = batch_encoding
            result.append(tokenized_document)
            for k, v in current_added_annotations.items():
                added_annotation_lists[k].extend(v)
            if added_annotations is not None:
                added_annotations.append(current_added_annotations)

    missed_annotations = defaultdict(set)
    if strict_span_conversion or verbose:
        # We check the annotations with respect to the layers of the result_document_type.
        # Note that the original document may have more layers, but since result documents
        # are of type result_document_type, we only check the layers of this type.
        for annotation_field in result_document_type.annotation_fields():
            # do not check the partition layer because the partitions are not required later on
            # and entries get quite probably removed when windowing is applied, so this just pollutes the logs
            if annotation_field.name != partition_layer:
                current_missed_annotations = set(doc[annotation_field.name]) - set(
                    added_annotation_lists[annotation_field.name]
                )
                if len(current_missed_annotations) > 0:
                    missed_annotations[annotation_field.name] = current_missed_annotations

    if len(missed_annotations) > 0:
        missed_annotations_simplified = {k: str(v) for k, v in missed_annotations.items()}
        if strict_span_conversion:
            raise ValueError(
                f"could not convert all annotations from document with id={doc.id} to token based documents, "
                f"but strict_span_conversion is True, so raise an error, "
                f"missed annotations:\n{json.dumps(missed_annotations_simplified, sort_keys=True, indent=2)}"
            )
        else:
            if verbose:
                logger.warning(
                    f"could not convert all annotations from document with id={doc.id} to token based documents, "
                    f"missed annotations (disable this message with verbose=False):\n"
                    f"{json.dumps(missed_annotations_simplified, sort_keys=True, indent=2)}"
                )

    return result


S = TypeVar("S", bound=Span)


class SpanNotAlignedWithTokenException(Exception):
    def __init__(self, span):
        self.span = span


def get_aligned_token_span(encoding: BatchEncoding, char_span: S) -> S:
    # find the start
    token_start = None
    token_end_before = None
    char_start = None
    for idx in range(char_span.start, char_span.end):
        token_start = encoding.char_to_token(idx)
        if token_start is not None:
            char_start = idx
            break

    if char_start is None:
        raise SpanNotAlignedWithTokenException(span=char_span)
    for idx in range(char_span.end - 1, char_start - 1, -1):
        token_end_before = encoding.char_to_token(idx)
        if token_end_before is not None:
            break

    if token_start is None or token_end_before is None:
        raise SpanNotAlignedWithTokenException(span=char_span)

    return char_span.copy(start=token_start, end=token_end_before + 1)



import logging
from collections.abc import Collection, Sized
from typing import Any, Callable, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar, Union

from torch import Tensor
from torchmetrics import Metric, MetricCollection
from torchmetrics.wrappers.abstract import WrapperMetric

logger = logging.getLogger(__name__)

T = TypeVar("T")
T2 = TypeVar("T2")


class WrappedMetricWithPrepareFunction(WrapperMetric, Generic[T]):
    """A wrapper around a metric that can be used with predictions and targets that are need to be
    prepared (e.g. un-batched) before passing them to the metric.

    Args:
        metric: The metric to wrap. It should be a subclass of torchmetrics.Metric.
        prepare_function: A function that prepares the input for the metric. If provided, It is called with
            the predictions as well as the targets (separately).
        prepare_together_function: A function that prepares both the predictions and the targets together and
            should return them as a tuple. If provided, it is called with the predictions and the targets as
            arguments.
        prepare_does_unbatch: If True, the prepare_function is expected to return an iterable of
            individual inputs. This can be used to un-batch the input before passing it to the
            wrapped metric.
    """

    def __init__(
        self,
        metric: Union[Metric, MetricCollection],
        prepare_function: Optional[Callable[[T], Any]] = None,
        prepare_together_function: Optional[Callable[[T, T], Tuple[Any, Any]]] = None,
        prepare_does_unbatch: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.metric = metric
        self.prepare_function = prepare_function
        self.prepare_both_function = prepare_together_function
        self.prepare_does_unbatch = prepare_does_unbatch

    def _is_empty_batch(self, prediction: T2, target: T2) -> bool:
        if isinstance(prediction, Sized) and isinstance(target, Sized):
            pred_len = len(prediction)
            target_len = len(target)
        else:
            raise ValueError(
                "Both prediction and target need to be sized when prepare_does_unbatch=False."
            )
        if pred_len != target_len:
            raise ValueError(
                f"Number of elements in prediction ({pred_len}) and target ({target_len}) do not match."
            )
        if pred_len == 0:
            return True
        return False

    def forward(self, prediction: T, target: T) -> Any:
        if self.prepare_function is not None:
            prediction = self.prepare_function(prediction)
            target = self.prepare_function(target)
        if self.prepare_both_function is not None:
            prediction, target = self.prepare_both_function(prediction, target)
        if self.prepare_does_unbatch:
            if not isinstance(prediction, Collection) or not isinstance(target, Collection):
                raise ValueError(
                    "Both prediction and target need to be iterable and sized when prepare_does_unbatch=True."
                )
            if len(prediction) != len(target):
                raise ValueError(
                    f"Number of prepared predictions ({len(prediction)}) and targets "
                    f"({len(target)}) do not match."
                )
            if len(prediction) == 0:
                raise ValueError("Empty batch.")
            results = []
            for prediction_str, target_str in zip(prediction, target):
                current_result = self.metric(prediction_str, target_str)
                results.append(current_result)
            return results
        else:
            if not self._is_empty_batch(prediction, target):
                return self.metric(prediction, target)
            else:
                return None

    def update(self, prediction: T, target: T) -> None:
        if self.prepare_function is not None:
            prediction = self.prepare_function(prediction)
            target = self.prepare_function(target)
        if self.prepare_both_function is not None:
            prediction, target = self.prepare_both_function(prediction, target)
        if self.prepare_does_unbatch:
            if not isinstance(prediction, Collection) or not isinstance(target, Collection):
                raise ValueError(
                    "Both prediction and target need to be iterable and sized when prepare_does_unbatch=True."
                )
            if len(prediction) != len(target):
                raise ValueError(
                    f"Number of prepared predictions ({len(prediction)}) and targets "
                    f"({len(target)}) do not match."
                )
            if len(prediction) == 0:
                raise ValueError("Empty batch.")
            for prediction_str, target_str in zip(prediction, target):
                self.metric.update(prediction_str, target_str)
        else:
            if not self._is_empty_batch(prediction, target):
                self.metric.update(prediction, target)

    def compute(self) -> Any:
        return self.metric.compute()

    def reset(self) -> None:
        self.metric.reset()

    @property
    def metric_state(self) -> Dict[str, Union[List[Tensor], Tensor]]:
        if isinstance(self.metric, Metric):
            return self.metric.metric_state
        elif isinstance(self.metric, MetricCollection):
            # TODO: maypy complains. is the code wrong or mypy?
            result = {
                metric_name: metric.metric_state for metric_name, metric in self.metric.items()
            }
            return result  # type: ignore[return-value]
        else:
            raise ValueError(f"Unsupported metric type: {type(self.metric)}")



import logging
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import torch
from pie_core import Model
from torch import FloatTensor, LongTensor, nn
from torch.nn import Parameter
from torch.optim import AdamW
from transformers import AutoConfig, AutoModel, PreTrainedModel, get_linear_schedule_with_warmup
from transformers.modeling_outputs import SequenceClassifierOutput
from typing_extensions import TypeAlias

from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

from .common import ModelWithBoilerplate
from .components.pooler import get_pooler_and_output_size

# model inputs / outputs / targets
InputType: TypeAlias = MutableMapping[str, LongTensor]
OutputType: TypeAlias = SequenceClassifierOutput
TargetType: TypeAlias = MutableMapping[str, Union[LongTensor, FloatTensor]]
# step inputs (batch) / outputs (loss)
StepInputType: TypeAlias = Tuple[InputType, TargetType]
StepOutputType: TypeAlias = FloatTensor


HF_MODEL_TYPE_TO_CLASSIFIER_DROPOUT_ATTRIBUTE = {
    "albert": "classifier_dropout_prob",
    "distilbert": "seq_classif_dropout",
}

logger = logging.getLogger(__name__)

T = TypeVar("T")


def separate_arguments_by_prefix(
    arguments: MutableMapping[str, T], prefixes: List[str]
) -> Dict[str, Dict[str, T]]:
    result: Dict[str, Dict[str, T]] = {prefix: {} for prefix in prefixes + ["remaining"]}
    for k, v in arguments.items():
        found = False
        for prefix in prefixes:
            if k.startswith(prefix):
                result[prefix][k[len(prefix) :]] = v
                found = True
                break
        if not found:
            result["remaining"][k] = v
    return result


class SequenceClassificationModelWithPoolerBase(
    ABC,
    ModelWithBoilerplate[InputType, OutputType, TargetType, StepOutputType],
    RequiresModelNameOrPath,
):
    """Abstract base model for sequence classification with a pooler.

    Args:
        model_name_or_path: The name or path of the HuggingFace model to use.
        tokenizer_vocab_size: The size of the tokenizer vocabulary. If provided, the model's
            tokenizer embeddings are resized to this size.
        classifier_dropout: The dropout probability for the classifier. If not provided, the
            dropout probability is taken from the Huggingface model config.
        learning_rate: The learning rate for the optimizer.
        task_learning_rate: The learning rate for the task-specific parameters. If None, the
            learning rate for all parameters is set to `learning_rate`.
        warmup_proportion: The proportion of steps to warm up the learning rate.
        pooler: The pooler configuration. If None, CLS token pooling is used.
        freeze_base_model: If True, the base model parameters are frozen.
        base_model_prefix: The prefix of the base model parameters when using a task_learning_rate
            or freeze_base_model. If None, the base_model_prefix of the model is used.
        **kwargs: Additional keyword arguments passed to the parent class,
            see :class:`ModelWithBoilerplate`.
    """

    def __init__(
        self,
        model_name_or_path: str,
        tokenizer_vocab_size: Optional[int] = None,
        classifier_dropout: Optional[float] = None,
        learning_rate: float = 1e-5,
        task_learning_rate: Optional[float] = None,
        warmup_proportion: float = 0.1,
        pooler: Optional[Union[Dict[str, Any], str]] = None,
        freeze_base_model: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.save_hyperparameters()

        self.learning_rate = learning_rate
        self.task_learning_rate = task_learning_rate
        self.warmup_proportion = warmup_proportion
        self.freeze_base_model = freeze_base_model
        self.model_name_or_path = model_name_or_path

        self.model = self.setup_base_model()

        if tokenizer_vocab_size is not None:
            self.model.resize_token_embeddings(tokenizer_vocab_size)

        if self.freeze_base_model:
            for param in self.model.parameters():
                param.requires_grad = False

        if classifier_dropout is None:
            # Get the classifier dropout value from the Huggingface model config.
            # This is a bit of a mess since some Configs use different variable names or change the semantics
            # of the dropout (e.g. DistilBert has one dropout prob for QA and one for Seq classification, and a
            # general one for embeddings, encoder and pooler).
            classifier_dropout_attr = HF_MODEL_TYPE_TO_CLASSIFIER_DROPOUT_ATTRIBUTE.get(
                self.model.config.model_type, "classifier_dropout"
            )
            classifier_dropout = getattr(self.model.config, classifier_dropout_attr) or 0.0
        self.dropout = nn.Dropout(classifier_dropout)

        if isinstance(pooler, str):
            pooler = {"type": pooler}
        self.pooler_config = pooler or {}
        self.pooler, pooler_output_dim = self.setup_pooler(input_dim=self.model.config.hidden_size)
        self.classifier = self.setup_classifier(pooler_output_dim=pooler_output_dim)
        self.loss_fct = self.setup_loss_fct()

    def setup_base_model(self) -> PreTrainedModel:
        config = AutoConfig.from_pretrained(self.model_name_or_path)
        if self.is_from_pretrained:
            return AutoModel.from_config(config=config)
        else:
            return AutoModel.from_pretrained(self.model_name_or_path, config=config)

    @abstractmethod
    def setup_classifier(self, pooler_output_dim: int) -> Callable:
        pass

    @abstractmethod
    def setup_loss_fct(self) -> Callable:
        pass

    def setup_pooler(self, input_dim: int) -> Tuple[Callable, int]:
        """Set up the pooler. The pooler is used to get a representation of the input sequence(s)
        that can be used by the classifier. It is a callable that takes the hidden states of the
        base model (and additional model inputs that are prefixed with "pooler_") and returns the
        pooled output.

        Args:
            input_dim: The input dimension of the pooler, i.e. the hidden size of the base model.

        Returns:
            A tuple with the pooler and the output dimension of the pooler.
        """
        return get_pooler_and_output_size(config=self.pooler_config, input_dim=input_dim)

    def get_pooled_output(self, model_inputs, pooler_inputs) -> torch.FloatTensor:
        output = self.model(**model_inputs)
        hidden_state = output.last_hidden_state
        pooled_output = self.pooler(hidden_state, **pooler_inputs)
        pooled_output = self.dropout(pooled_output)
        return pooled_output

    def forward(
        self,
        inputs: InputType,
        targets: Optional[TargetType] = None,
        return_hidden_states: bool = False,
    ) -> OutputType:
        sanitized_inputs = separate_arguments_by_prefix(arguments=inputs, prefixes=["pooler_"])

        pooled_output = self.get_pooled_output(
            model_inputs=sanitized_inputs["remaining"], pooler_inputs=sanitized_inputs["pooler_"]
        )

        logits = self.classifier(pooled_output)

        result = {"logits": logits}
        if targets is not None:
            labels = targets["labels"]
            loss = self.loss_fct(logits, labels)
            result["loss"] = loss
        if return_hidden_states:
            raise NotImplementedError("return_hidden_states is not yet implemented")

        return SequenceClassifierOutput(**result)

    @abstractmethod
    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        pass

    def base_model_named_parameters(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        if prefix:
            prefix = f"{prefix}."
        return self.model.named_parameters(prefix=f"{prefix}model")

    def task_named_parameters(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        if prefix:
            prefix = f"{prefix}."
        base_model_parameter_names = dict(self.base_model_named_parameters(prefix=prefix)).keys()
        for name, param in self.named_parameters(prefix=prefix):
            if name not in base_model_parameter_names:
                yield name, param

    def configure_optimizers(self):
        if self.task_learning_rate is not None:
            base_model_params = (param for name, param in self.base_model_named_parameters())
            task_params = (param for name, param in self.task_named_parameters())
            optimizer = AdamW(
                [
                    {"params": base_model_params, "lr": self.learning_rate},
                    {"params": task_params, "lr": self.task_learning_rate},
                ]
            )
        else:
            optimizer = AdamW(self.parameters(), lr=self.learning_rate)

        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer


@Model.register()
class SequenceClassificationModelWithPooler(
    SequenceClassificationModelWithPoolerBase,
    RequiresNumClasses,
):
    """A sequence classification model that uses a pooler to get a representation of the input
    sequence and then applies a linear classifier to that representation. The pooler can be
    configured via the `pooler` argument, see :func:`get_pooler_and_output_size` for details.

    Args:
        num_classes: The number of classes for the classification task.
        multi_label: If True, the model is trained as a multi-label classifier.
        multi_label_threshold: The threshold for the multi-label classifier, i.e. the probability
            above which a class is predicted.
        **kwargs
    """

    def __init__(
        self,
        num_classes: int,
        multi_label: bool = False,
        multi_label_threshold: float = 0.5,
        **kwargs,
    ):
        # set num_classes and multi_label before call to super init because they are used there
        # in setup_classifier and setup_loss_fct
        self.num_classes = num_classes
        self.multi_label = multi_label
        super().__init__(**kwargs)

        self.multi_label_threshold = multi_label_threshold

    def setup_classifier(self, pooler_output_dim: int) -> Callable:
        return nn.Linear(pooler_output_dim, self.num_classes)

    def setup_loss_fct(self) -> Callable:
        return nn.BCEWithLogitsLoss() if self.multi_label else nn.CrossEntropyLoss()

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        if not self.multi_label:
            labels = torch.argmax(outputs.logits, dim=-1).to(torch.long)
            probabilities = torch.softmax(outputs.logits, dim=-1)
        else:
            probabilities = torch.sigmoid(outputs.logits)
            labels = (probabilities > self.multi_label_threshold).to(torch.long)
        assert isinstance(probabilities, FloatTensor)
        assert isinstance(labels, LongTensor)
        return {"labels": labels, "probabilities": probabilities}


@Model.register()
class SequencePairSimilarityModelWithPooler(
    SequenceClassificationModelWithPoolerBase,
):
    """A span pair similarity model to detect of two spans occurring in different texts are
    similar. It uses an encoder to independently calculate contextualized embeddings of both texts,
    then uses a pooler to get representations of the spans and, finally, calculates the cosine to
    get the similarity scores.

    Args:
        label_threshold: The threshold above which score the spans are considered as similar.
        pooler: The pooler identifier or config, see :func:`get_pooler_and_output_size` for details.
            Defaults to "mention_pooling" (max pooling over the span token embeddings).
        **kwargs
    """

    def __init__(
        self,
        pooler: Optional[Union[Dict[str, Any], str]] = None,
        **kwargs,
    ):
        if pooler is None:
            # use (max) mention pooling per default
            pooler = {"type": "mention_pooling", "num_indices": 1}
        super().__init__(pooler=pooler, **kwargs)

    def setup_classifier(
        self, pooler_output_dim: int
    ) -> Callable[[torch.FloatTensor, torch.FloatTensor], torch.FloatTensor]:
        return torch.nn.functional.cosine_similarity  # type: ignore[return-value]

    def setup_loss_fct(self) -> Callable:
        return nn.BCELoss()

    def forward(
        self,
        inputs: InputType,
        targets: Optional[TargetType] = None,
        return_hidden_states: bool = False,
    ) -> OutputType:
        sanitized_inputs = separate_arguments_by_prefix(
            # Note that the order of the prefixes is important because one is a prefix of the other,
            # so we need to start with the longer!
            arguments=inputs,
            prefixes=["pooler_pair_", "pooler_"],
        )

        pooled_output = self.get_pooled_output(
            model_inputs=sanitized_inputs["remaining"]["encoding"],
            pooler_inputs=sanitized_inputs["pooler_"],
        )
        pooled_output_pair = self.get_pooled_output(
            model_inputs=sanitized_inputs["remaining"]["encoding_pair"],
            pooler_inputs=sanitized_inputs["pooler_pair_"],
        )

        logits = self.classifier(pooled_output, pooled_output_pair)

        result = {"logits": logits}
        if targets is not None:
            labels = targets["scores"]
            loss = self.loss_fct(logits, labels)
            result["loss"] = loss
        if return_hidden_states:
            raise NotImplementedError("return_hidden_states is not yet implemented")

        return SequenceClassifierOutput(**result)

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        # probabilities = torch.sigmoid(outputs.logits)
        scores = outputs.logits
        return {"scores": scores}
